# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the OpenAITransformer."""

from __future__ import annotations

from time import sleep
from typing import Any, ClassVar, Dict, List, Optional, Type

import openai  # pylint: disable=import-error
from autotransform.batcher.base import Batch
from autotransform.command.base import FACTORY as command_factory
from autotransform.command.base import Command
from autotransform.config import get_config
from autotransform.event.handler import EventHandler
from autotransform.event.verbose import VerboseEvent
from autotransform.item.base import Item
from autotransform.item.file import FileItem
from autotransform.transformer.base import TransformerName
from autotransform.transformer.single import SingleTransformer
from autotransform.validator.base import FACTORY as validator_factory
from autotransform.validator.base import ValidationResultLevel, Validator
from pydantic import Field, validator


class OpenAITransformer(SingleTransformer):
    """A transformer which uses OpenAI models to perform a completion to generate code.

    Attributes:
        prompt (str): The prompt to use for completition. Uses sentry values to replace
            values in the prompt.
            <<FILE_PATH>> - Replaced with the path of the file being transformed.
            <<FILE_CONTENT>> - Replaced with the content of the file being transformed.
        commands (optional, List[Command]): A set of commands to use on transformed files
            before validation. Useful for correcting things like formatting. Defaults to an
            empty list.
        max_completion_attempts (optional, int): The maximum number of times to try the OpenAI
            API. Defaults to 3.
        max_validation_attempts (optional, int): The maximum number of times to validate
            completitions. Defaults to 3.
        model (optional, str): The model to use for completition. Defaults to gpt-3.5-turbo.
        system_message (optional, Optional[str]): The system message to use. Defaults to None.
        temperature (optional, float): The temperature to use to control the quality of outputs.
            Defaults to 0.4.
        validators (optional, List[Validator]): A set of validators to use to provide feedback to
            the LLM with issues it may have produced. Defaults to an empty list.
        name (ClassVar[TransformerName]): The name of the component.
    """

    prompt: str
    commands: List[Command] = Field(default_factory=list)
    max_completion_attempts: int = 3
    max_validation_attempts: int = 3
    model: str = "gpt-3.5-turbo"
    system_message: Optional[str] = None
    temperature: float = 0.4
    validators: List[Validator] = Field(default_factory=list)

    name: ClassVar[TransformerName] = TransformerName.OPEN_AI

    # pylint: disable=invalid-name
    @validator("max_completion_attempts")
    @classmethod
    def max_completion_attempts_must_be_positive(cls, v: int) -> int:
        """Validates the max_completion_attempts is a positive number.

        Args:
            v (int): The maximum number of completion attempts for the OpenAI API.

        Raises:
            ValueError: Raises an error when the max_completion_attempts is not positive.

        Returns:
            int: The unmodified max_completion_attempts.
        """

        if v < 1:
            raise ValueError("The max completion attempts must be at least 1")
        return v

    # pylint: disable=invalid-name
    @validator("max_validation_attempts")
    @classmethod
    def max_validation_attempts_must_be_positive(cls, v: int) -> int:
        """Validates the max_validation_attempts is a positive number.

        Args:
            v (int): The maximum number of validation attempts for completion.

        Raises:
            ValueError: Raises an error when the max_validation_attempts is not positive.

        Returns:
            int: The unmodified max_validation_attempts.
        """

        if v < 1:
            raise ValueError("The max validation attempts must be at least 1")
        return v

    # pylint: disable=invalid-name
    @validator("temperature")
    @classmethod
    def temperature_must_be_valid(cls, v: float) -> float:
        """Validates the temperature is between 0 and 1.

        Args:
            v (float): The temperature to use for model completion.

        Raises:
            ValueError: Raises an error when the temperature is not in the valid range.

        Returns:
            int: The unmodified temperature.
        """

        if v >= 1.0 or v <= 0.0:
            raise ValueError("The temperature must be between 0.0 and 1.0")
        return v

    def _transform_item(self, item: Item) -> None:
        """Replaces a file with the completition results from an OpenAI completition.

        Args:
            item (Item): The file that will be transformed.
        """

        if openai.api_key is None:
            openai.api_key = get_config().open_ai_api_key

        assert isinstance(item, FileItem)

        event_handler = EventHandler.get()
        original_content = item.get_content()
        batch: Batch = {"title": "test", "items": [item]}

        # Set up messages for prompt
        messages = []
        if self.system_message:
            messages.append({"role": "system", "content": self.system_message})
        messages.append(
            {
                "role": "user",
                "content": self._replace_sentinel_values(self.prompt, item),
            }
        )

        completion_success = False
        for _ in range(self.max_validation_attempts):
            # Get completion
            chat_completion = None
            for i in range(self.max_completion_attempts):
                try:
                    chat_completion = openai.ChatCompletion.create(
                        model=self.model,
                        messages=messages,
                        temperature=self.temperature,
                    )
                    break
                except Exception as e:  # pylint: disable=broad-exception-caught
                    chat_completion = None
                    sleep(min(4 ** (i+1), 60))
                    event_handler.handle(
                        VerboseEvent({"message": f"API Failure on {item.get_path()}: {e}"}),
                    )

            if chat_completion is None:
                item.write_content(original_content)
                return

            # Log completion information
            token_usage = (
                f"Prompt: {chat_completion.usage.prompt_tokens}"
                + f" - Completition: {chat_completion.usage.completion_tokens}"
            )
            event_handler.handle(VerboseEvent({"message": token_usage}))
            completition_result = chat_completion.choices[0].message.content
            message = f"The completion result for {item.get_path()}:\n\n{completition_result}"
            event_handler.handle(VerboseEvent({"message": message}))

            # Update File
            item.write_content(self._extract_code_from_completion(completition_result))

            # Run commands to fix file
            for command in self.commands:
                try:
                    command.run(batch, None)
                except Exception:  # pylint: disable=broad-exception-caught
                    event_handler.handle(
                        VerboseEvent({"message": f"Failed to run command {command}"})
                    )

            # Run validators to identify issues with completion
            failures = []
            for completion_validator in self.validators:
                validation_result = completion_validator.check(batch, None)
                if validation_result.level != ValidationResultLevel.NONE:
                    failures.append(str(validation_result.message))

            # Check if another completion is required
            completion_success = not failures
            if completion_success:
                break

            # Add messages for handling failures for next completion
            messages.append({"role": "assistant", "content": completition_result})
            failure_message = "\n".join(failures)
            messages.append(
                {
                    "role": "user",
                    "content": f"The following errors were found\n{failure_message}"
                    + "provide the file with fixes for these errors.",
                },
            )

        # If we had validation failures on our last run, just use the original content
        if not completion_success:
            event_handler.handle(
                VerboseEvent({"message": "Completion failed, using original content"})
            )
            item.write_content(original_content)

    def _replace_sentinel_values(self, prompt: str, item: FileItem) -> str:
        """Replaces sentinel values in a prompt

        Args:
            prompt (str): The prompt with potential sentinel values to replace.
            item (FileItem): The Item to use for replacing sentinel values.

        Returns:
            str: The prompt with sentinel values replaced.
        """
        new_prompt = prompt.replace("<<FILE_PATH>>", item.get_path())
        return new_prompt.replace("<<FILE_CONTENT>>", item.get_content())

    def _extract_code_from_completion(self, result: str) -> str:
        """Extracts code from the result of an OpenAI completition.

        Args:
            result (str): The completition result.

        Returns:
            str: The extracted code from the completition result.
        """

        code_lines = []
        in_code = False
        # Checks for code inside formatting blocks
        for line in result.split("\n"):
            if line.startswith("```"):
                if in_code:
                    break
                in_code = True
                continue
            if in_code:
                code_lines.append(line)
        # Hit when no formatting is present or only trailing backticks
        if not in_code or not code_lines:
            code = result.removesuffix("```")
        else:
            code = "\n".join(code_lines)
        return code

    @classmethod
    def from_data(cls: Type[OpenAITransformer], data: Dict[str, Any]) -> OpenAITransformer:
        """Produces an instance of the component from decoded data.

        Args:
            data (Mapping[str, Any]): The JSON decoded data.

        Returns:
            OpenAITransformer: An instance of the component.
        """

        prompt = data["prompt"]
        commands = [command_factory.get_instance(c) for c in data.get("commands", [])]
        model = data.get("model", "gpt-3.5-turbo")
        system_message = data.get("system_message", None)
        temperature = data.get("temperature", 0.4)
        validators = [validator_factory.get_instance(v) for v in data.get("validators", [])]

        return cls(
            prompt=prompt,
            commands=commands,
            model=model,
            system_message=system_message,
            temperature=temperature,
            validators=validators,
        )



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Validators take in a Batch that has undergone transformation and validate that the
changes do not produce any issues/errors. They may run test cases, check linting, perform
type checks, or some other process."""



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The VerboseEvent is a simple, generic event used for logging information to
the console. This information is not expected to be logged to long term storage and
is used to follow usage in production environments. More specific events should be created
for long term storage cases.
"""

from typing import TypedDict

from autotransform.event.base import Event
from autotransform.event.logginglevel import LoggingLevel
from autotransform.event.type import EventType


class VerboseEventData(TypedDict):
    """The data for a VerboseEvent. Contains the information that will be
    logged when the event is triggered."""

    message: str


class VerboseEvent(Event[VerboseEventData]):
    """A simple, generic event used to log verbose information to the console."""

    @staticmethod
    def get_type() -> EventType:
        """Used to represent the type of Event, output to logs.

        Returns:
            EventType: The unique type associated with this Event.
        """

        return EventType.VERBOSE

    @staticmethod
    def get_logging_level() -> LoggingLevel:
        """The logging level for events of this type.

        Returns:
            LoggingLevel: The logging detail required to log this event.
        """

        return LoggingLevel.VERBOSE

    def _get_message(self) -> str:
        """Gets a message representing the details of the event.

        Returns:
            str: The message for the event.
        """

        return self.data["message"]



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the GithubChange."""

from __future__ import annotations

import json
import re
from functools import cached_property
from typing import ClassVar, Dict, List, Tuple

from autotransform.batcher.base import Batch
from autotransform.change.base import Change, ChangeName, ChangeState, ReviewState, TestState
from autotransform.config import get_config
from autotransform.item.base import FACTORY as item_factory
from autotransform.schema.schema import AutoTransformSchema
from autotransform.util.github import GithubUtils, PullRequest
from autotransform.util.schema_map import SchemaMap


class GithubChange(Change):
    """A Change representing a Pull Request on a Github repo.

    Attributes:
        full_github_name (str): The fully qualified name of the Github Repo the
            Pull Request is against.
        pull_number (int): The number for the Pull Request.
        name (ClassVar[ChangeName]): The name of the Component.
    """

    full_github_name: str
    pull_number: int
    name: ClassVar[ChangeName] = ChangeName.GITHUB

    def get_batch(self) -> Batch:
        """Gets the Batch that was used to produce the Change.

        Returns:
            Batch: The Batch used to produce the Change.
        """

        return self._automation_data[1]

    def get_schema(self) -> AutoTransformSchema:
        """Gets the Schema that was used to produce the Change.

        Returns:
            AutoTransformSchema: The Schema used to produce the Change.
        """

        schema_name = self.get_schema_name()
        schema = SchemaMap.get().get_schema(schema_name)
        assert schema_name == schema.config.schema_name
        repo_override = get_config().repo_override
        if repo_override is not None:
            schema.repo = repo_override

        return schema

    def get_schema_name(self) -> str:
        """Gets the name of the Schema that produced the Change.

        Returns:
            str: The name of the Schema.
        """

        return self._automation_data[0].config.schema_name

    def get_state(self) -> ChangeState:
        """Gets the current state of the Change.

        Returns:
            ChangeState: The current state of the Change.
        """

        if self._pull_request.merged:
            return ChangeState.MERGED
        if self._pull_request.is_closed():
            return ChangeState.CLOSED
        return ChangeState.OPEN

    def get_mergeable_state(self) -> str:
        """Gets the mergeable state of the Change.

        Returns:
            ChangeState: The mergeable state of the Change.
        """

        return self._pull_request.get_mergeable_state()

    def get_review_state(self) -> ReviewState:
        """Gets the current review state of the Change.

        Returns:
            ReviewState: The current review state of the Change.
        """

        return self._review_state

    @cached_property
    def _review_state(self) -> ReviewState:
        """The current review state of the Change as a cached property.

        Returns:
            ReviewState: The current review state of the Change.
        """

        review_state = self._pull_request.get_review_state()
        if review_state == "APPROVED":
            return ReviewState.APPROVED
        if review_state == "CHANGES_REQUESTED":
            return ReviewState.CHANGES_REQUESTED
        return ReviewState.NEEDS_REVIEW

    def get_test_state(self) -> TestState:
        """Gets the current test state of the Change.

        Returns:
            TestState: The current test state of the Change.
        """

        return self._test_state

    @cached_property
    def _test_state(self) -> TestState:
        """The current test state of the Change as a cached property.

        Returns:
            TestState: The current test state of the Change.
        """

        state = self._pull_request.get_test_state()
        if state == "pending":
            return TestState.PENDING
        if state in ["success", "neutral"]:
            return TestState.SUCCESS
        return TestState.FAILURE

    def get_labels(self) -> List[str]:
        """Gets all labels for a Change.

        Returns:
            List[str]: The list of labels.
        """

        return self._pull_request.get_labels()

    def get_reviewers(self) -> List[str]:
        """Gets all reviewers for a Change.

        Returns:
            List[str]: The list of reviewers.
        """

        return self._pull_request.get_reviewers()

    def get_team_reviewers(self) -> List[str]:
        """Gets all team reviewers for a Change.

        Returns:
            List[str]: The list of team reviewers.
        """

        return self._pull_request.get_team_reviewers()

    def get_created_timestamp(self) -> int:
        """Returns the timestamp when the Pull Request was created.

        Returns:
            int: The timestamp in seconds when the Pull Request was created.
        """

        return self._pull_request.get_created_at()

    def get_last_updated_timestamp(self) -> int:
        """Returns the timestamp when the Pull Request was last updated.

        Returns:
            int: The timestamp in seconds when the Pull Request was last updated.
        """

        return self._pull_request.get_updated_at()

    def abandon(self) -> bool:
        """Close the Pull Request and delete the associated branch.

        Returns:
            bool: Whether the abandon was completed successfully.
        """

        if not self._pull_request.close():
            return False
        return self._pull_request.delete_branch()

    def add_labels(self, labels: List[str]) -> bool:
        """Adds labels to an outstanding Change.

        Args:
            labels (List[str]): The labels to add.

        Returns:
            bool: Whether the labels were added successfully.
        """

        self._pull_request.add_labels(labels)
        return True

    def add_reviewers(self, reviewers: List[str], team_reviewers: List[str]) -> bool:
        """Adds reviewers to an outstanding Change.

        Args:
            reviewers (List[str]): The reviewers to add.
            team_reviewers (List[str]): Any team reviewers to add.

        Returns:
            bool: Whether the reviewers were added successfully.
        """

        self._pull_request.add_reviewers(reviewers, team_reviewers)
        return True

    def comment(self, body: str) -> bool:
        """Comments on an outstanding Change.

        Args:
            body (str): The body of the comment.

        Returns:
            bool: Whether the comment was successful.
        """

        self._pull_request.create_comment(body)
        return True

    def merge(self) -> bool:
        """Merges the Pull Request and deletes the branch.

        Returns:
            bool: Whether the merge was completed successfully.
        """

        if not self._pull_request.merge():
            return False
        return self._pull_request.delete_branch()

    def remove_label(self, label: str) -> bool:
        """Removes a label from an outstanding Change.

        Args:
            label (str): The label to remove.

        Returns:
            bool: Whether the label was removed successfully.
        """

        self._pull_request.remove_label(label)
        return True

    @property
    def _pull_request(self) -> PullRequest:
        """Gets the Pull Request.

        Returns:
            PullRequest: The Pull Request.
        """

        return GithubUtils.get(self.full_github_name).get_pull_request(self.pull_number)

    @cached_property
    def _automation_data(self) -> Tuple[AutoTransformSchema, Batch]:
        """Loads the Schema and Batch data for the GithubChange as a cached property.

        Returns:
            Tuple[AutoTransformSchema, Batch]: The Schema and Batch contained
                in the PullRequest's Body.
        """

        gist_match = re.search("<<<Automation Info Gist: (.*)>>>", self._pull_request.body)
        if gist_match:
            gist_ids = gist_match.groups()[0].split("/")
            gist = GithubUtils.get(self.full_github_name).get_gist(gist_ids[0])
            schema_data = gist.get_file_content("schema") or ""
            batch = json.loads(gist.get_file_content("batch") or "")
            assert isinstance(batch["items"], List)
            for i in range(1, len(gist_ids)):
                gist = GithubUtils.get(self.full_github_name).get_gist(gist_ids[i])
                items = json.loads(gist.get_file_content("items") or "[]")
                assert isinstance(items, List)
                batch["items"].extend(items)
        else:
            cur_line_placement = None
            data_lines: Dict[str, List[str]] = {"schema": [], "batch": []}
            for line in self._pull_request.body.splitlines():
                if line == GithubUtils.BEGIN_SCHEMA:
                    cur_line_placement = "schema"
                elif line == GithubUtils.END_SCHEMA:
                    cur_line_placement = None
                elif line == GithubUtils.BEGIN_BATCH:
                    cur_line_placement = "batch"
                elif line == GithubUtils.END_BATCH:
                    cur_line_placement = None
                elif cur_line_placement is not None:
                    data_lines[cur_line_placement].append(line)
            schema_data = "\n".join(data_lines["schema"])
            batch = json.loads("\n".join(data_lines["batch"]))

        schema = AutoTransformSchema.from_data(json.loads(schema_data))
        items = [item_factory.get_instance(item) for item in batch["items"]]
        batch = {
            "items": items,
            "metadata": batch["metadata"],
            "title": str(batch["title"]),
        }
        return (schema, batch)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the RegexBatcher."""

from __future__ import annotations

import re
from typing import Any, ClassVar, Dict, List, Sequence

from pydantic import Field

from autotransform.batcher.base import Batch, Batcher, BatcherName
from autotransform.item.base import Item
from autotransform.item.file import FileItem


class FileRegexBatcher(Batcher):
    """A Batcher which uses matches from regex on file content to group Items.

    Attributes:
        group_by (str): The regex which produces the group by value.
        metadata_keys (optional, Dict[str, str]): A mapping from key to a regex that produces values
            for that key.
        name (ClassVar[BatcherName]): The name of the Component.
    """

    group_by: str
    metadata_keys: Dict[str, str] = Field(default_factory=dict)

    name: ClassVar[BatcherName] = BatcherName.FILE_REGEX

    def batch(self, items: Sequence[Item]) -> List[Batch]:
        """Take filtered Items and group them by regex match values.

        Args:
            items (Sequence[Item]): The filtered Items to separate.

        Returns:
            List[Batch]: A list of Batches grouped by the extra_data of the Items.
        """

        groups: Dict[str, List[FileItem]] = {}
        for item in items:
            assert isinstance(item, FileItem)
            match = re.match(item.get_content(), self.group_by)
            assert match is not None, "Must have value to use for grouping"
            group_by_val = match.group(1)
            if group_by_val in groups:
                groups[group_by_val].append(item)
            else:
                groups[group_by_val] = [item]

        batches: List[Batch] = []
        for group_title, group_items in groups.items():
            batch: Batch = {"items": group_items, "title": group_title}
            if self.metadata_keys:
                metadata: Dict[str, List[Any]] = {}
                for key in self.metadata_keys:
                    metadata[key] = []
                for item in group_items:
                    file_content = item.get_content()
                    for key, regex in self.metadata_keys.items():
                        match = re.match(file_content, regex)
                        if match:
                            metadata[key].append(match.group(1))
                for key in self.metadata_keys:
                    metadata[key] = list(set(metadata[key]))
                batch["metadata"] = metadata
            batches.append(batch)
        return batches



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Schemas are the heart of AutoTransform. They contain all necesary components and configuration
to deploy a change.
"""

# pylint: disable=invalid-name

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from autotransform.schema.schema import AutoTransformSchema

# A variable to store the currently executing schema so that components can access it without
# having to thread it.
current: Optional[AutoTransformSchema] = None



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Batcher components."""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import Any, ClassVar, List, Mapping, Sequence, TypedDict

from typing_extensions import NotRequired

from autotransform.item.base import Item
from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent


class BatcherName(str, Enum):
    """A simple enum for mapping."""

    CHUNK = "chunk"
    CODEOWNERS = "codeowners"
    DIRECTORY = "directory"
    EXTRA_DATA = "extra_data"
    FILE_REGEX = "file_regex"
    SINGLE = "single"


class Batch(TypedDict):
    """A logical grouping of Items with title and associated metadata."""

    items: Sequence[Item]
    metadata: NotRequired[Mapping[str, Any]]
    title: str


class Batcher(NamedComponent):
    """The base for Batcher components. Used by AutoTransform to separate Items in to logical
    groupings that can be acted on indepently and have associated metadata.

    Attributes:
        name (ClassVar[BatcherName]): The name of the Component.
    """

    name: ClassVar[BatcherName]

    @abstractmethod
    def batch(self, items: Sequence[Item]) -> List[Batch]:
        """Take filtered Items and separate in to logical groupings with associated group metadata
        and title.

        Args:
            items (Sequence[Item]): The filtered Items to separate.

        Returns:
            List[Batch]: A list of logical groupings of Items with associated group metadata and
                title.
        """


FACTORY = ComponentFactory(
    {
        BatcherName.CHUNK: ComponentImport(
            class_name="ChunkBatcher", module="autotransform.batcher.chunk"
        ),
        BatcherName.CODEOWNERS: ComponentImport(
            class_name="CodeownersBatcher", module="autotransform.batcher.codeowners"
        ),
        BatcherName.DIRECTORY: ComponentImport(
            class_name="DirectoryBatcher", module="autotransform.batcher.directory"
        ),
        BatcherName.EXTRA_DATA: ComponentImport(
            class_name="ExtraDataBatcher", module="autotransform.batcher.extradata"
        ),
        BatcherName.FILE_REGEX: ComponentImport(
            class_name="FileRegexBatcher", module="autotransform.batcher.regex"
        ),
        BatcherName.SINGLE: ComponentImport(
            class_name="SingleBatcher", module="autotransform.batcher.single"
        ),
    },
    Batcher,  # type: ignore [type-abstract]
    "batcher.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The Config represents settings that control how AutoTransform operates."""

from __future__ import annotations

import json
import os
from typing import Any, Dict, Optional, Tuple

from autotransform.repo.base import FACTORY as repo_factory
from autotransform.repo.base import Repo
from autotransform.runner.base import FACTORY as runner_factory
from autotransform.runner.base import Runner
from autotransform.runner.github import GithubRunner
from autotransform.runner.jenkins import JenkinsAPIRunner
from autotransform.runner.local import LocalRunner
from autotransform.util.component import ComponentModel
from autotransform.util.console import choose_yes_or_no, get_str
from pydantic import Field


class Config(ComponentModel):
    """A collection of settings for configuring the functionality of AutoTransform.

    Attributes:
        component_directory (Optional[str], optional): The directory where custom component
            JSON files are located. If not provided, autotransform/ will be used.
            Defaults to None.
        github_token (Optional[str], optional): The Github token to use for authentication.
            Defaults to None.
        github_base_url (Optional[str], optional): The base URL for API requests to Github.
            Used for Github Enterprise. Defaults to None.
        jenkins_user (Optional[str], optional): The Jenkins username for authentication. Defaults
            to None.
        jenkins_token (Optional[str], optional): The Jenkins token to use for authentication.
            Defaults to None.
        jenkins_base_url (Optional[str], optional): The base URL for requests to Jenkins.
            Defaults to None.
        local_runner (Optional[Runner], optional): The Runner to use for local runs.
            Defaults to None.
        open_ai_api_key (Optional[str], optional): The API key to use for OpenAI completitions.
            Defaults to None.
        remote_runner (Optional[Runner], optional): The runner to use for remote runs.
            Defaults to None.
        repo_override (Optional[Repo], optional): A repo to use in place of any schema repos.
            Defaults to None.
    """

    component_directory: Optional[str] = None
    github_token: Optional[str] = Field(default=None, redact=True)
    github_base_url: Optional[str] = None
    jenkins_user: Optional[str] = None
    jenkins_token: Optional[str] = Field(default=None, redact=True)
    jenkins_base_url: Optional[str] = None
    local_runner: Optional[Runner] = None
    open_ai_api_key: Optional[str] = None
    remote_runner: Optional[Runner] = None
    repo_override: Optional[Repo] = None

    def write(self, file_path: str) -> None:
        """Writes the Config to a file as JSON.

        Args:
            file_path (str): The file to write the Config to.
        """

        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w+", encoding="UTF-8") as schedule_file:
            schedule_file.write(json.dumps(self.bundle(), indent=4))
            schedule_file.flush()

    @staticmethod
    def read(file_path: str) -> Config:
        """Reads the Config from a JSON encoded file.

        Args:
            file_path (str): The path where the JSON for the Config is located.

        Returns:
            Config: The Config from the file.
        """

        try:
            with open(file_path, "r", encoding="UTF-8") as config_file:
                config_json = config_file.read()
            return Config.from_json(config_json)
        except FileNotFoundError:
            return Config()

    @staticmethod
    def from_json(config_json: str) -> Config:
        """Builds a Config from JSON encoded values.

        Args:
            config_json (str): The JSON encoded Config.

        Returns:
            Config: The Config from the JSON.
        """

        return Config.from_data(json.loads(config_json))

    @staticmethod
    def from_data(data: Dict[str, Any]) -> Config:
        """Produces an instance of the Config from decoded data.

        Args:
            data (Mapping[str, Any]): The JSON decoded data.

        Returns:
            Config: An instance of the Config.
        """

        github_token = data.get("github_token", None)
        if github_token is not None:
            assert isinstance(github_token, str)

        github_base_url = data.get("github_base_url", None)
        if github_base_url is not None:
            assert isinstance(github_base_url, str)

        jenkins_user = data.get("jenkins_user", None)
        if jenkins_user is not None:
            assert isinstance(jenkins_user, str)

        jenkins_token = data.get("jenkins_token", None)
        if jenkins_token is not None:
            assert isinstance(jenkins_token, str)

        jenkins_base_url = data.get("jenkins_base_url", None)
        if jenkins_base_url is not None:
            assert isinstance(jenkins_base_url, str)

        open_ai_api_key = data.get("open_ai_api_key", None)
        if open_ai_api_key is not None:
            assert isinstance(open_ai_api_key, str)

        component_directory = data.get("component_directory", None)
        if component_directory is not None:
            assert isinstance(component_directory, str)

        local_runner = data.get("local_runner", None)
        if local_runner is not None:
            local_runner = runner_factory.get_instance(local_runner)

        remote_runner = data.get("remote_runner", None)
        if remote_runner is not None:
            remote_runner = runner_factory.get_instance(remote_runner)

        repo_override = data.get("repo_override", None)
        if repo_override is not None:
            repo_override = repo_factory.get_instance(repo_override)

        return Config(
            github_token=github_token,
            github_base_url=github_base_url,
            jenkins_user=jenkins_user,
            jenkins_token=jenkins_token,
            jenkins_base_url=jenkins_base_url,
            component_directory=component_directory,
            local_runner=local_runner,
            open_ai_api_key=open_ai_api_key,
            remote_runner=remote_runner,
            repo_override=repo_override,
        )

    @staticmethod
    def get_github_token_from_console(
        prev_config: Optional[Config] = None,
        simple: bool = False,
        user_config: bool = False,
    ) -> Optional[str]:
        """Gets the Github token using console inputs.

        Args:
            prev_config (Optional[Config], optional): Previously input Config. Defaults to None.
            simple (bool, optional): Whether to use the simple setup. Defaults to False.
            user_config (bool, optional): Whether this configuration is for a user level Config.
                Defaults to False.

        Returns:
            Optional[str]: The Github token.
        """

        if not user_config:
            return None
        if prev_config is not None and (simple or choose_yes_or_no("Use previous Github Token?")):
            return prev_config.github_token

        github_token = get_str(
            "Enter the Github token to use(empty to not include one): ", secret=True
        )
        if github_token in ["", "None"]:
            return None

        return github_token

    @staticmethod
    def get_github_base_url_from_console(
        prev_config: Optional[Config] = None,
        simple: bool = False,
    ) -> Optional[str]:
        """Gets the Github base URL using console inputs.

        Args:
            prev_config (Optional[Config], optional): Previously input Config. Defaults to None.
            simple (bool, optional): Whether to use the simple setup. Defaults to False.

        Returns:
            Optional[str]: The Github base URL.
        """

        # Potentially use previous GHE information
        if prev_config is not None and (
            simple
            or choose_yes_or_no(f"Use previous Github base url: {prev_config.github_base_url}?")
        ):
            return prev_config.github_base_url

        # Simple setups assume no Github Enterprise
        if simple:
            return None

        github_base_url = get_str(
            "Enter the base URL for Github API requests (i.e. https://api.<org>-github.com): "
        )
        if github_base_url in ["", "None"]:
            return None
        return github_base_url

    @staticmethod
    def get_jenkins_user_from_console(
        prev_config: Optional[Config] = None,
        simple: bool = False,
    ) -> Optional[str]:
        """Gets the Jenkins user using console inputs.

        Args:
            prev_config (Optional[Config], optional): Previously input Config. Defaults to None.
            simple (bool, optional): Whether to use the simple setup. Defaults to False.

        Returns:
            Optional[str]: The Jenkins user.
        """

        # Potentially use previous Jenkins information
        if prev_config is not None and (
            simple or choose_yes_or_no(f"Use previous Jenkins user: {prev_config.jenkins_user}?")
        ):
            return prev_config.jenkins_user

        # Simple setups assume no Jenkins
        if simple:
            return None

        jenkins_user = get_str("Enter the Jenkins user: ")
        if jenkins_user in ["", "None"]:
            return None
        return jenkins_user

    @staticmethod
    def get_jenkins_token_from_console(
        prev_config: Optional[Config] = None,
        simple: bool = False,
        user_config: bool = False,
    ) -> Optional[str]:
        """Gets the Jenkins token using console inputs.

        Args:
            prev_config (Optional[Config], optional): Previously input Config. Defaults to None.
            simple (bool, optional): Whether to use the simple setup. Defaults to False.
            user_config (bool, optional): Whether this configuration is for a user level Config.
                Defaults to False.

        Returns:
            Optional[str]: The Jenkins token.
        """

        if not user_config:
            return None
        if prev_config is not None and (simple or choose_yes_or_no("Use previous Jenkins Token?")):
            return prev_config.jenkins_token

        jenkins_token = get_str(
            "Enter the Jenkins token to use(empty to not include one): ", secret=True
        )
        if jenkins_token in ["", "None"]:
            return None

        return jenkins_token

    @staticmethod
    def get_jenkins_base_url_from_console(
        prev_config: Optional[Config] = None,
        simple: bool = False,
    ) -> Optional[str]:
        """Gets the Jenkins base URL using console inputs.

        Args:
            prev_config (Optional[Config], optional): Previously input Config. Defaults to None.
            simple (bool, optional): Whether to use the simple setup. Defaults to False.

        Returns:
            Optional[str]: The Jenkins base URL.
        """

        # Potentially use previous Jenkins information
        if prev_config is not None and (
            simple
            or choose_yes_or_no(f"Use previous Jenkins base url: {prev_config.jenkins_base_url}?")
        ):
            return prev_config.jenkins_base_url

        # Simple setups assume no Jenkins
        if simple:
            return None

        jenkins_base_url = get_str(
            "Enter the base URL for Jenkins requests (i.e. https://<org>.jenkins.com): "
        )
        if jenkins_base_url in ["", "None"]:
            return None
        return jenkins_base_url

    @staticmethod
    def get_open_ai_api_key_from_console(
        prev_config: Optional[Config] = None,
        simple: bool = False,
        user_config: bool = False,
    ) -> Optional[str]:
        """Gets the OpenAI API key using console inputs.

        Args:
            prev_config (Optional[Config], optional): Previously input Config. Defaults to None.
            simple (bool, optional): Whether to use the simple setup. Defaults to False.
            user_config (bool, optional): Whether this configuration is for a user level Config.
                Defaults to False.

        Returns:
            Optional[str]: The Jenkins token.
        """

        if not user_config:
            return None
        if prev_config is not None and (simple or choose_yes_or_no("Use previous OpenAI API Key?")):
            return prev_config.open_ai_api_key

        open_ai_api_key = get_str(
            "Enter the OpenAI API key(empty to not include one): ", secret=True
        )
        if open_ai_api_key in ["", "None"]:
            return None

        return open_ai_api_key

    @staticmethod
    def get_component_directory_from_console(
        prev_config: Optional[Config] = None,
        simple: bool = False,
    ) -> Optional[str]:
        """Gets the component directory using console inputs.

        Args:
            prev_config (Optional[Config], optional): Previously input Config. Defaults to None.
            simple (bool, optional): Whether to use the simple setup. Defaults to False.

        Returns:
            Optional[str]: The component directory.
        """

        # Potentially use previous information
        if prev_config is not None and (
            simple
            or choose_yes_or_no(
                f"Use previous component directory: {prev_config.component_directory}?"
            )
        ):
            return prev_config.component_directory

        # Simple setups assume no custom components
        if simple:
            return None

        component_directory = get_str("Enter the directory with custom component JSON files: ")
        if component_directory in ["", "None"]:
            return None
        return component_directory

    @staticmethod
    def get_local_runner_from_console(
        prev_config: Optional[Config] = None,
        simple: bool = False,
    ) -> Optional[Runner]:
        """Gets the local runner using console inputs.

        Args:
            prev_config (Optional[Config], optional): Previously input Config. Defaults to None.
            simple (bool, optional): Whether to use the simple setup. Defaults to False.

        Returns:
            Optional[Runner]: The local runner.
        """

        default_runner = LocalRunner()

        args: Dict[str, Any] = {
            "default_value": default_runner,
            "simple": simple,
        }
        if prev_config is not None:
            args["previous_value"] = prev_config.local_runner
        return runner_factory.from_console("local runner", **args)

    @staticmethod
    def get_remote_runner_from_console(
        prev_config: Optional[Config] = None,
        use_github: bool = False,
        use_jenkins: bool = False,
        simple: bool = False,
    ) -> Optional[Runner]:
        """Gets the remote runner using console inputs.

        Args:
            prev_config (Optional[Config], optional): Previously input Config. Defaults to None.
            use_github (bool, optional): Whether to use Github or not. Defaults to False.
            use_jenkins (bool, optional): Whether to use Jenkins or not. Defaults to False.
            simple (bool, optional): Whether to use the simple setup. Defaults to False.

        Returns:
            Optional[Runner]: The local runner.
        """

        args: Dict[str, Any] = {
            "simple": simple,
        }
        if use_jenkins:
            args["default_value"] = JenkinsAPIRunner(job_name="autotransform")
        elif use_github:
            args["default_value"] = GithubRunner(
                run_workflow="autotransform.run.yml", update_workflow="autotransform.update.yml"
            )

        if prev_config is not None:
            args["previous_value"] = prev_config.remote_runner
        return runner_factory.from_console("remote runner", **args)

    @staticmethod
    def get_repo_override_from_console(
        prev_config: Optional[Config] = None,
        simple: bool = False,
    ) -> Optional[Repo]:
        """Gets the repo override using console inputs.

        Args:
            prev_config (Optional[Config], optional): Previously input Config. Defaults to None.
            simple (bool, optional): Whether to use the simple setup. Defaults to False.

        Returns:
            Optional[Repo]: The repo override.
        """

        args: Dict[str, Any] = {
            "simple": simple,
        }
        if prev_config is not None:
            args["previous_value"] = prev_config.repo_override

        return repo_factory.from_console("repo override", **args)

    @staticmethod
    def from_console(
        prev_config: Optional[Config] = None,
        simple: bool = False,
        use_github: Optional[bool] = None,
        use_jenkins: Optional[bool] = None,
        user_config: bool = False,
    ) -> Tuple[Config, bool, bool]:
        """Gets a Config using console inputs.

        Args:
            prev_config (Optional[Config], optional): Previously input Config. Defaults to None.
            simple (bool, optional): Whether to use the simple setup. Defaults to False.
            use_github (bool, optional): Whether to use Github or not. Defaults to None.
            use_jenkins (bool, optional): Whether to use Jenkins or not. Defaults to None.
            user_config (bool, optional): Whether this configuration is for a user level Config.
                Defaults to False.

        Returns:
            Tuple[Config, bool, bool]: The input Config and whether it uses Github and Jenkins.
        """

        # Determine whether to include Github settings
        if use_github:
            github = True
        elif use_github is not None:
            github = False
        else:
            github = choose_yes_or_no("Use Github for this config?")

        if not github:
            github_token = None
            github_base_url = None
        else:
            github_token = Config.get_github_token_from_console(
                prev_config=prev_config, simple=simple, user_config=user_config
            )
            github_base_url = Config.get_github_base_url_from_console(
                prev_config=prev_config, simple=simple
            )

        if use_jenkins:
            jenkins = True
        elif use_jenkins is not None:
            jenkins = False
        else:
            jenkins = choose_yes_or_no("Use Jenkins for this config?")

        if not jenkins:
            jenkins_user = None
            jenkins_token = None
            jenkins_base_url = None
        else:
            jenkins_user = Config.get_jenkins_user_from_console(
                prev_config=prev_config, simple=simple
            )
            jenkins_token = Config.get_jenkins_token_from_console(
                prev_config=prev_config, simple=simple, user_config=user_config
            )
            jenkins_base_url = Config.get_jenkins_base_url_from_console(
                prev_config=prev_config, simple=simple
            )

        return (
            Config(
                github_token=github_token,
                github_base_url=github_base_url,
                jenkins_user=jenkins_user,
                jenkins_token=jenkins_token,
                jenkins_base_url=jenkins_base_url,
                component_directory=Config.get_component_directory_from_console(
                    prev_config=prev_config, simple=simple
                ),
                local_runner=Config.get_local_runner_from_console(
                    prev_config=prev_config, simple=simple
                ),
                open_ai_api_key=Config.get_open_ai_api_key_from_console(
                    prev_config=prev_config, simple=simple,
                ),
                remote_runner=Config.get_remote_runner_from_console(
                    prev_config=prev_config, use_github=github, use_jenkins=jenkins, simple=simple
                ),
                repo_override=Config.get_repo_override_from_console(
                    prev_config=prev_config, simple=simple
                ),
            ),
            github,
            jenkins,
        )

    def merge(self, other: Config) -> Config:
        """Merges the Config with another Config. Settings in the other Config will override
        those in this Config.

        Args:
            other (Config): The Config to be merged with.

        Returns:
            Config: The merged Config.
        """

        return Config(
            github_token=other.github_token or self.github_token,
            github_base_url=other.github_base_url or self.github_base_url,
            jenkins_user=other.jenkins_user or self.jenkins_user,
            jenkins_token=other.jenkins_token or self.jenkins_token,
            jenkins_base_url=other.jenkins_base_url or self.jenkins_base_url,
            component_directory=other.component_directory or self.component_directory,
            local_runner=other.local_runner or self.local_runner,
            open_ai_api_key=other.open_ai_api_key or self.open_ai_api_key,
            remote_runner=other.remote_runner or self.remote_runner,
            repo_override=other.repo_override or self.repo_override,
        )



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The AutoTransform script, responsible for handling AutoTransform CLI invocations.
Different CLI commands are handled as subparsers."""

from argparse import ArgumentParser

from autotransform.scripts.commands import initialize, manage, run, schedule, settings, update


def get_arg_parser() -> ArgumentParser:
    """Gets the argument parser for AutoTransform. Sets up each command as a sub parser.

    Returns:
        ArgumentParser: The arg parser with all args set up.
    """

    parser = ArgumentParser(
        description="AutoTransform is a tool for structured, automated code modifcation",
        prog="AutoTransform",
    )
    subparsers = parser.add_subparsers()

    # Add run command
    run_parser = subparsers.add_parser(
        "run",
        help="Execute a full run of AutoTransform",
    )
    run.add_args(run_parser)

    # Add schedule command
    schedule_parser = subparsers.add_parser(
        "schedule",
        help="Schedule runs of AutoTransform schemas",
    )
    schedule.add_args(schedule_parser)

    # Add manage command
    manage_parser = subparsers.add_parser(
        "manage",
        help="Manage outstanding Changes",
    )
    manage.add_args(manage_parser)

    # Add update command
    update_parser = subparsers.add_parser(
        "update",
        help="Update an outstanding change",
    )
    update.add_args(update_parser)

    # Add settings command
    settings_parser = subparsers.add_parser(
        "settings",
        help="Update/view AutoTransform settings",
    )
    settings.add_args(settings_parser)

    # Add initialize command
    init_parser = subparsers.add_parser(
        "init",
        help="Initialize AutoTransform for the user/repo",
    )
    initialize.add_args(init_parser)

    return parser


def main():
    """Parse the arguments of a script run and execute the command invoked."""

    parser = get_arg_parser()
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the ScriptCommand."""

from __future__ import annotations

from typing import Any, ClassVar, List, Mapping, Optional, Sequence

import autotransform.schema
from autotransform.batcher.base import Batch
from autotransform.command.base import Command, CommandName
from autotransform.event.handler import EventHandler
from autotransform.event.verbose import VerboseEvent
from autotransform.item.base import Item
from autotransform.item.file import FileItem
from autotransform.util.functions import run_cmd_on_items


class ScriptCommand(Command):
    """Runs a script with the supplied arguments to perform a command. If the per_item flag is
    set, the script will be invoked on each Item. If run_on_changes is set to True, the script
    will replace the Batch Items with FileItems for each changed file. Sentinel values can be
    used in args to provide custom arguments for a run.
    The available sentinel values for args are:
        <<KEY>>: A json encoded list of the Items for a Batch. If the per_item flag is set
            this will simply be the key of an Item.
        <<EXTRA_DATA>>: A JSON encoded mapping from Item key to that Item's extra_data. If the
            per_item flag is set, this will simply be a JSON encoding of the Item's extra_data.
            If extra_data is not present for an item, it is treated as an empty Dict.
        <<METADATA>>: A JSON encoded version of the Batch's metadata.
    _FILE can be appended to any of these (i.e. <<KEY_FILE>>) and the arg will instead be replaced
     with a path to a file containing the value.

    Attributes:
        args (List[str]): The arguments to supply to the script.
        script (str): The script to run.
        per_item (bool, optional): Whether to run the script on each item. Defaults to False.
        run_on_changes (bool, optional): Whether to replace the Items in the batch with
            FileItems for the changed files. Defaults to False.
        run_pre_validation (bool, optional): Whether to run the command before validation is done.
            Defaults to False.
        name (ClassVar[CommandName]): The name of the Component.
    """

    args: List[str]
    script: str
    per_item: bool = False
    run_on_changes: bool = False
    run_pre_validation: bool = False

    name: ClassVar[CommandName] = CommandName.SCRIPT

    def run(self, batch: Batch, _transform_data: Optional[Mapping[str, Any]]) -> None:
        """Runs the script command against the Batch, either on each item individually or
        on the entire Batch, based on the per_item flag.

        Args:
            batch (Batch): The transformed Batch to run against.
            _transform_data (Optional[Mapping[str, Any]]): Data from the transformation. Unused.
        """

        if not self.per_item:
            self._run_batch(batch)
            return
        if self.run_on_changes:
            current_schema = autotransform.schema.current
            assert current_schema is not None
            repo = current_schema.repo
            assert repo is not None
            items: Sequence[Item] = [FileItem(key=file) for file in repo.get_changed_files(batch)]
        else:
            items = batch["items"]

        for item in items:
            self._run_single(item, batch.get("metadata", None))

    def _run_single(self, item: Item, batch_metadata: Optional[Mapping[str, Any]]) -> None:
        """Executes a simple script to run a command on a single Item.

        Args:
            item (Item): The Item that will be validated.
            batch_metadata (Optional[Mapping[str, Any]]): The metadata of the Batch containing the
                Item.
        """

        event_handler = EventHandler.get()

        # Get Command
        cmd = [self.script]
        cmd.extend(self.args)

        proc = run_cmd_on_items(cmd, [item], batch_metadata or {})

        # Handle output
        if proc.stdout.strip() != "":
            event_handler.handle(VerboseEvent({"message": f"STDOUT:\n{proc.stdout.strip()}"}))
        else:
            event_handler.handle(VerboseEvent({"message": "No STDOUT"}))
        if proc.stderr.strip() != "":
            event_handler.handle(VerboseEvent({"message": f"STDERR:\n{proc.stderr.strip()}"}))
        else:
            event_handler.handle(VerboseEvent({"message": "No STDERR"}))
        proc.check_returncode()

    def _run_batch(self, batch: Batch) -> None:
        """Executes a simple script against the given Batch.

        Args:
            batch (Batch): The batch that will be run against.
        """

        event_handler = EventHandler.get()

        # Get items
        if self.run_on_changes:
            current_schema = autotransform.schema.current
            assert current_schema is not None
            repo = current_schema.repo
            assert repo is not None
            items: Sequence[Item] = [FileItem(key=file) for file in repo.get_changed_files(batch)]
            if len(items) == 0:
                return
        else:
            items = batch["items"]

        # Get Command
        cmd = [self.script]
        cmd.extend(self.args)

        proc = run_cmd_on_items(cmd, items, batch.get("metadata", {}))

        # Handle output
        if proc.stdout.strip() != "":
            event_handler.handle(VerboseEvent({"message": f"STDOUT:\n{proc.stdout.strip()}"}))
        else:
            event_handler.handle(VerboseEvent({"message": "No STDOUT"}))
        if proc.stderr.strip() != "":
            event_handler.handle(VerboseEvent({"message": f"STDERR:\n{proc.stderr.strip()}"}))
        else:
            event_handler.handle(VerboseEvent({"message": "No STDERR"}))
        proc.check_returncode()



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Command components."""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import Any, ClassVar, Mapping, Optional

from autotransform.batcher.base import Batch
from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent


class CommandName(str, Enum):
    """A simple enum for mapping."""

    SCRIPT = "script"


class Command(NamedComponent):
    """The base for Command components. Used by AutoTransform to perform post-processing
    operations after validation, such as code generation.

    Attributes:
        run_pre_validation (bool, optional): Whether to run the command before validation.
            Defaults to False.
        name (ClassVar[CommandName]): The name of the Component.
    """

    run_pre_validation: bool = False

    name: ClassVar[CommandName]

    @abstractmethod
    def run(self, batch: Batch, transform_data: Optional[Mapping[str, Any]]) -> None:
        """Performs the post-processing steps represented by the Command.

        Args:
            batch (Batch): The Batch for which this Command is run.
            transform_data (Optional[Mapping[str, Any]]): Data from the transformation.
        """


FACTORY = ComponentFactory(
    {
        CommandName.SCRIPT: ComponentImport(
            class_name="ScriptCommand", module="autotransform.command.script"
        ),
    },
    Command,  # type: ignore [type-abstract]
    "command.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the ChunkBatcher."""

from __future__ import annotations

from copy import deepcopy
from math import ceil
from typing import Any, ClassVar, Dict, List, Optional, Sequence

from autotransform.batcher.base import Batch, Batcher, BatcherName
from autotransform.item.base import Item


class ChunkBatcher(Batcher):
    """A batcher which chunks Items in to several Batches. A chunk size is supplied that
    determines the size of Batches. A maximum number of chunks can optionally be supplied. If
    the chunk size would result in more than the maximum number of chunks, the chunk size is
    increased to the minimum required to create only max_chunks. The title for a batch will
    have [X/N] prepended to it, where X represents the chunk number and N represents the
    total number of chunks.

    Attributes:
        chunk_size (int): The size of chunks.
        title (str): The title to use for Batches.
        max_chunks (Optional[int], optional): The maximum number of chunks to create. Defaults
            to None.
        metadata (Optional[Dict[str, Any]], optional): The metadata to associate with
            Batches. Defaults to None.
        name (ClassVar[BatcherName]): The name of the Component.
    """

    chunk_size: int
    title: str
    max_chunks: Optional[int] = None
    metadata: Optional[Dict[str, Any]] = None

    name: ClassVar[BatcherName] = BatcherName.CHUNK

    def batch(self, items: Sequence[Item]) -> List[Batch]:
        """Take filtered Items and chunk them into Batches.

        Args:
            items (Sequence[Item]): The filtered Items to separate.

        Returns:
            List[Batch]: A list of Batches representing each chunk of the filtered
                Items.
        """

        # Ensure no more than max chunks
        chunk_size = self.chunk_size
        if self.max_chunks is not None and len(items) / chunk_size > self.max_chunks:
            chunk_size = ceil(len(items) / self.max_chunks)

        # Create Batches
        item_chunks = [items[i : i + chunk_size] for i in range(0, len(items), chunk_size)]
        item_batches: List[Batch] = []
        for idx, item_chunk in enumerate(item_chunks, start=1):
            title = f"[{idx}/{len(item_chunks)}] " + self.title
            batch: Batch = {"items": item_chunk, "title": title}
            if self.metadata is not None:
                # Deepcopy metadata to ensure mutations don't apply to all Batches
                batch["metadata"] = deepcopy(self.metadata)
            item_batches.append(batch)
        return item_batches



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The manage command is used to manage outstanding Changes for AutoTransform."""

from argparse import ArgumentParser, Namespace

from autotransform.config import get_repo_config_relative_path
from autotransform.event.handler import EventHandler
from autotransform.event.logginglevel import LoggingLevel
from autotransform.event.run import ScriptRunEvent
from autotransform.event.verbose import VerboseEvent
from autotransform.util.manager import Manager


def add_args(parser: ArgumentParser) -> None:
    """Adds the args to a subparser that are required to manage outstanding changes.

    Args:
        parser (ArgumentParser): The parser for the managing outstanding changes.
    """

    parser.add_argument(
        "--path",
        metavar="path",
        required=False,
        type=str,
        help="A file path to the JSON encoded file, only use if file is in a non-usual place.",
    )

    logging_level = parser.add_mutually_exclusive_group()
    logging_level.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        required=False,
        help="Tells the script to output verbose logs.",
    )
    logging_level.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        required=False,
        help="Tells the script to output debug logs.",
    )

    mode_group = parser.add_mutually_exclusive_group()
    mode_group.add_argument(
        "-l",
        "--local",
        dest="run_local",
        action="store_true",
        required=False,
        help="Tells the script to use the local runner.",
    )
    mode_group.add_argument(
        "-r",
        "--remote",
        dest="run_local",
        action="store_false",
        required=False,
        help="Tells the script to use the remote runner. This is the default mode.",
    )

    parser.set_defaults(run_local=False, func=manage_command_main)


def manage_command_main(args: Namespace) -> None:
    """The main method for the manage command, handles the actual management of changes.

    Args:
        args (Namespace): The arguments supplied to the manage command, such as the JSON file.
    """

    # pylint: disable=unspecified-encoding

    event_handler = EventHandler.get()
    if args.verbose:
        event_handler.set_logging_level(LoggingLevel.VERBOSE)
    if args.debug:
        event_handler.set_logging_level(LoggingLevel.DEBUG)

    manager_file = args.path
    if manager_file is None:
        manager_file = f"{get_repo_config_relative_path()}/manager.json"
    event_args = {"manager_file": manager_file}
    manager = Manager.read(manager_file)
    event_args["manager"] = manager
    event_handler.handle(ScriptRunEvent({"script": "manage", "args": event_args}))

    event_handler.handle(VerboseEvent({"message": f"Running manager: {manager!r}"}))
    manager.run(args.run_local)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the KeyHashShardFilter."""

from __future__ import annotations

from hashlib import md5
from typing import ClassVar

from autotransform.filter.base import FilterName
from autotransform.filter.shard import ShardFilter
from autotransform.item.base import Item


class KeyHashShardFilter(ShardFilter):
    """A Filter which produces a shard from the key of an Item using md5 hashing.

    Attributes:
        name (ClassVar[FilterName]): The name of the component.
    """

    name: ClassVar[FilterName] = FilterName.KEY_HASH_SHARD

    def _shard(self, item: Item) -> int:
        """Produces a shard from an item.

        Args:
            item (Item): The Item to produce a shard number for.

        Returns:
            int: The shard number for the Item.
        """

        return int(md5(item.key.encode("UTF-8")).hexdigest(), 16) % self.num_shards



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Validator components."""

from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, ClassVar, Dict, Mapping, Optional

from autotransform.batcher.base import Batch
from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent


class ValidationResultLevel(str, Enum):
    """The result level of a validation indicating how bad a validation failure is."""

    NONE = "none"
    WARNING = "warning"
    ERROR = "error"

    def compare(self, other: str) -> int:
        """Compares two result levels and returns an integer indicating the comparison.

        Args:
            other (str): The value to compare against.

        Returns:
            int: A negative number if lt, 0 if equal, and a positive number if gt.
        """

        if self.value is other:
            return 0
        val_order: Dict[str, int] = {"none": 0, "warning": 1, "error": 2}
        return val_order[self.value] - val_order[other]

    def __eq__(self, other: object) -> bool:
        return self.compare(str(other.value) if isinstance(other, Enum) else str(other)) == 0

    def __ne__(self, other: object) -> bool:
        return self.compare(str(other.value) if isinstance(other, Enum) else str(other)) != 0

    def __lt__(self, other: object) -> bool:
        return self.compare(str(other.value) if isinstance(other, Enum) else str(other)) < 0

    def __le__(self, other: object) -> bool:
        return self.compare(str(other.value) if isinstance(other, Enum) else str(other)) <= 0

    def __gt__(self, other: object) -> bool:
        return self.compare(str(other.value) if isinstance(other, Enum) else str(other)) > 0

    def __ge__(self, other: object) -> bool:
        return self.compare(str(other.value) if isinstance(other, Enum) else str(other)) >= 0


@dataclass(frozen=True, kw_only=True)
class ValidationResult:
    """Represents the result of an attempt at validation.

    Attributes:
        level (ValidationResultLevel): The level of the validation issue raised.
        validator (Validator): The Validator that returned this result.
        message (Optional[str], optional): The message associated with the validation
            result. Defaults to None.
    """

    level: ValidationResultLevel
    validator: Validator
    message: Optional[str] = None


@dataclass(frozen=True, kw_only=True)
class ValidationError(Exception):
    """An error raised by validation failing on a run.

    Attributes:
        issue (ValidationResult): The validation result that triggered the error.
        message (str): A message representing why the validation failed.
    """

    issue: ValidationResult
    message: Optional[str]


class ValidatorName(str, Enum):
    """A simple enum for mapping."""

    SCRIPT = "script"


class Validator(NamedComponent):
    """The base for Validator components. Validators test that the codebase is still
    healthy after a transformation.

    Attributes:
        name (ClassVar[ValidatorName]): The name of the component.
    """

    name: ClassVar[ValidatorName]

    @abstractmethod
    def check(self, batch: Batch, transform_data: Optional[Mapping[str, Any]]) -> ValidationResult:
        """Validate that a Batch that has undergone transformation does not produce any issues
        such as test failures or type errors.

        Args:
            batch (Batch): The transformed Batch to validate.
            transform_data (Optional[Mapping[str, Any]]): Data from the transformation.

        Returns:
            ValidationResult: The result of the validation check indicating the severity of any
                validation failures as well as an associated message.
        """


FACTORY = ComponentFactory(
    {
        ValidatorName.SCRIPT: ComponentImport(
            class_name="ScriptValidator", module="autotransform.validator.script"
        ),
    },
    Validator,  # type: ignore [type-abstract]
    "validator.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the JenkinsRunner."""

from __future__ import annotations

import json
import os
from typing import Any, ClassVar, Dict

import requests
from autotransform.change.base import Change
from autotransform.config import get_config
from autotransform.event.handler import EventHandler
from autotransform.event.verbose import VerboseEvent
from autotransform.event.warning import WarningEvent
from autotransform.filter.shard import ShardFilter
from autotransform.runner.base import Runner, RunnerName
from autotransform.schema.schema import AutoTransformSchema


class JenkinsAPIRunner(Runner):
    """A Runner component that uses Jenkins API requests for remote runs.

    Attributes:
        job_name (str): The name of the Jenkins job.
        name (ClassVar[RunnerName]): The name of the component.
    """

    job_name: str

    name: ClassVar[RunnerName] = RunnerName.JENKINS_API

    def run(self, schema: AutoTransformSchema) -> None:
        """Triggers a full run of a Schema using a Jenkins API request.

        Args:
            schema (AutoTransformSchema): The schema that will be run.
        """

        job_params = {"COMMAND": "run", "SCHEMA_NAME": schema.config.schema_name}
        if schema.config.max_submissions:
            job_params["MAX_SUBMISSIONS"] = str(schema.config.max_submissions)
        shard_filter = [filt for filt in schema.filters if isinstance(filt, ShardFilter)]
        if shard_filter:
            job_params["FILTER"] = json.dumps(shard_filter[0].bundle())

        self._run_jenkins_job(self.job_name, job_params)

    def update(self, change: Change) -> None:
        """Triggers an update of the Change using a Jenkins API request.

        Args:
            change (Change): The Change to update.
        """

        self._run_jenkins_job(
            self.job_name, {"COMMAND": "update", "CHANGE": json.dumps(change.bundle())}
        )

    @staticmethod
    def _run_jenkins_job(job_name: str, params: Dict[str, Any]) -> None:
        """Runs a Jenkins job to handle a run/update of AutoTransform.

        Args:
            job_name (str): The name of the Jenkins job to trigger.
            params (Dict[str, Any]): The params to pass to the Jenkins job.
        """

        config = get_config()
        event_handler = EventHandler.get()
        jenkins_user = config.jenkins_user
        jenkins_token = config.jenkins_token
        if jenkins_user is None or jenkins_token is None:
            event_handler.handle(
                WarningEvent({"message": "User and token must be provided to use Jenkins"})
            )
            return

        try:
            auth = (jenkins_user, jenkins_token)
            crumb_data = requests.get(
                f"{config.jenkins_base_url}/crumbIssuer/api/json",
                auth=auth,
                headers={"content-type": "application/json"},
                timeout=120,
            )
            if str(crumb_data.status_code) == "200":
                data = requests.get(
                    f"{config.jenkins_base_url}/job/{job_name}/buildWithParameters",
                    auth=auth,
                    params=params,
                    headers={
                        "content-type": "application/json",
                        "Jenkins-Crumb": crumb_data.json()["crumb"],
                    },
                    timeout=120,
                )

                if str(data.status_code) == "201":
                    event_handler.handle(VerboseEvent({"message": "Jenkins job is triggered"}))
                else:
                    event_handler.handle(
                        WarningEvent({"message": "Failed to trigger the Jenkins job"})
                    )

            else:
                event_handler.handle(WarningEvent({"message": "Couldn't fetch Jenkins-Crumb"}))

        # pylint: disable=broad-except
        except Exception as ex:
            event_handler.handle(WarningEvent({"message": "Failed triggering the Jenkins job"}))
            event_handler.handle(WarningEvent({"message": f"Error: {str(ex)}"}))


class JenkinsFileRunner(Runner):
    """A Runner component that creates files to trigger Jenkins jobs
    using https://plugins.jenkins.io/parameterized-trigger/.

    Attributes:
        name (ClassVar[RunnerName]): The name of the component.
        num_files (ClassVar[int]): The number of files created by the runner.
    """

    name: ClassVar[RunnerName] = RunnerName.JENKINS_FILE
    num_files: ClassVar[int] = 0

    def run(self, schema: AutoTransformSchema) -> None:
        """Triggers a full run of a Schema by creating a file with the appropriate content.

        Args:
            schema (AutoTransformSchema): The schema that will be run.
        """

        job_params = {"COMMAND": "run", "SCHEMA_NAME": schema.config.schema_name}
        if schema.config.max_submissions:
            job_params["MAX_SUBMISSIONS"] = str(schema.config.max_submissions)
        shard_filter = [filt for filt in schema.filters if isinstance(filt, ShardFilter)]
        if shard_filter:
            job_params["FILTER"] = json.dumps(shard_filter[0].bundle())

        self._create_file(job_params)

    def update(self, change: Change) -> None:
        """Triggers an update of the Change using a Jenkins API request.

        Args:
            change (Change): The Change to update.
        """

        self._create_file({"COMMAND": "update", "CHANGE": json.dumps(change.bundle())})

    def _create_file(self, props: Dict[str, str]) -> None:
        """Creates the file for the Jenkins job.

        Args:
            props (Dict[str, str]): The props to add to the job.
        """

        JenkinsFileRunner.num_files += 1

        cwd = os.getcwd().replace("\\", "/")
        file_path = f"{cwd}/autotransform/jenkins/job_{JenkinsFileRunner.num_files}.txt"

        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        content = "\n".join([f"{k}={v}" for k, v in props.items()])
        with open(file_path, "w+", encoding="UTF-8") as job_file:
            job_file.write(content)
            job_file.flush()



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""A ConfigFetcher that uses config files for configuration."""

from autotransform.config import CONFIG_FILE_NAME, get_cwd_config_dir, get_repo_config_dir
from autotransform.config.config import Config
from autotransform.config.fetcher import ConfigFetcher
from autotransform.util.package import get_config_dir


# pylint: disable=too-few-public-methods
class DefaultConfigFetcher(ConfigFetcher):
    """The default configuration fetcher that pulls from the config files. Three possible
    files are used:
     - a file within the AutoTransform package itself, located at autotransform-config/config.json
     - a file within the repo, located at {repo_root}/autotransform/config.json
     - a file relative to the current working directory, located at {cwd}/autotransform/config.json
    Settings from files later in the list will override settings from files earlier in the list. CWD
    configs will have the greatest priority, followed by repo configs, then the overall config.

    """

    def get_config(self) -> Config:
        """Fetch the Config.

        Returns:
            Config: The Config for AutoTransform.
        """

        potential_paths = [
            get_config_dir(),
            get_repo_config_dir(),
            get_cwd_config_dir(),
        ]
        config_paths = [
            f"{path}/{CONFIG_FILE_NAME}" for path in potential_paths if path is not None
        ]
        config = Config()
        for path in config_paths:
            config = config.merge(Config.read(path))
        return config



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the EmptyInput."""

from __future__ import annotations

from typing import ClassVar, Sequence

from autotransform.input.base import Input, InputName
from autotransform.item.base import Item


class EmptyInput(Input):
    """An Input that simply returns an empty list. Used when a Transformer operates
    on the whole codebase, rather than on an individual Item/set of Items.

    Attributes:
        name (ClassVar[InputName]): The name of the component.
    """

    name: ClassVar[InputName] = InputName.EMPTY

    def get_items(self) -> Sequence[Item]:
        """Returns an empty list of Items, useful for Transformers that operate
        on the whole codebase at once.

        Returns:
            Sequence[Item]: An empty list of Items.
        """
        return []



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Action components."""

from abc import abstractmethod
from enum import Enum
from typing import ClassVar

from autotransform.change.base import Change
from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent


class ActionName(str, Enum):
    """A simple enum for mapping."""

    ABANDON = "abandon"
    ADD_LABELS = "add_labels"
    ADD_OWNERS_AS_REVIEWERS = "add_owners_as_reviewers"
    ADD_OWNERS_AS_TEAM_REVIEWERS = "add_owners_as_team_reviewers"
    ADD_REVIEWERS = "add_reviewers"
    COMMENT = "comment"
    MERGE = "merge"
    NONE = "none"
    REMOVE_LABEL = "remove_label"
    REQUEST = "request"
    UPDATE = "update"


class Action(NamedComponent):
    """The base for Action components. Used by AutoTransform to perform some task on
    outstanding Changes based on Steps in the Manager. Tasks include options such as
    updating, abandoning, merging, and more.

    Attributes:
        name (ClassVar[ActionName]): The name of the component.
    """

    name: ClassVar[ActionName]

    @abstractmethod
    def run(self, change: Change) -> bool:
        """Performs the action on the specified Change.

        Args:
            change (Change): The Change to perform the Action on.

        Returns:
            bool: Whether the Action was successful.
        """


FACTORY = ComponentFactory(
    {
        ActionName.ABANDON: ComponentImport(
            class_name="AbandonAction", module="autotransform.step.action.source"
        ),
        ActionName.ADD_LABELS: ComponentImport(
            class_name="AddLabelsAction", module="autotransform.step.action.labels"
        ),
        ActionName.ADD_REVIEWERS: ComponentImport(
            class_name="AddReviewersAction", module="autotransform.step.action.reviewers"
        ),
        ActionName.ADD_OWNERS_AS_REVIEWERS: ComponentImport(
            class_name="AddOwnersAsReviewersAction", module="autotransform.step.action.reviewers"
        ),
        ActionName.ADD_OWNERS_AS_TEAM_REVIEWERS: ComponentImport(
            class_name="AddOwnersAsTeamReviewersAction",
            module="autotransform.step.action.reviewers",
        ),
        ActionName.COMMENT: ComponentImport(
            class_name="CommentAction", module="autotransform.step.action.comments"
        ),
        ActionName.MERGE: ComponentImport(
            class_name="MergeAction", module="autotransform.step.action.source"
        ),
        ActionName.NONE: ComponentImport(
            class_name="NoneAction", module="autotransform.step.action.source"
        ),
        ActionName.REMOVE_LABEL: ComponentImport(
            class_name="RemoveLabelAction", module="autotransform.step.action.labels"
        ),
        ActionName.REQUEST: ComponentImport(
            class_name="RequestAction", module="autotransform.step.action.request"
        ),
        ActionName.UPDATE: ComponentImport(
            class_name="UpdateAction", module="autotransform.step.action.source"
        ),
    },
    Action,  # type: ignore [type-abstract]
    "action.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The update command is used to update an outstanding Change."""

import json
import os
from argparse import ArgumentParser, Namespace

from autotransform.change.base import FACTORY as change_factory
from autotransform.config import get_config
from autotransform.event.debug import DebugEvent
from autotransform.event.handler import EventHandler
from autotransform.event.logginglevel import LoggingLevel
from autotransform.event.run import ScriptRunEvent
from autotransform.event.verbose import VerboseEvent
from autotransform.runner.base import Runner
from autotransform.runner.local import LocalRunner


def add_args(parser: ArgumentParser) -> None:
    """Adds the args to a subparser that are required to update a change.

    Args:
        parser (ArgumentParser): The parser for the change update.
    """

    parser.add_argument(
        "change",
        metavar="change",
        type=str,
        help="The change that will be updated. Could be a file path, string, "
        + "or environment variable name.",
    )

    logging_level = parser.add_mutually_exclusive_group()
    logging_level.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        required=False,
        help="Tells the script to output verbose logs.",
    )
    logging_level.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        required=False,
        help="Tells the script to output debug logs.",
    )

    type_group = parser.add_mutually_exclusive_group()
    type_group.add_argument(
        "-f",
        "--file",
        dest="change_type",
        action="store_const",
        const="file",
        required=False,
        help="Tells the script to interpret the change as a file path.",
    )
    type_group.add_argument(
        "-s",
        "--string",
        dest="change_type",
        action="store_const",
        const="string",
        required=False,
        help="Tells the script to interpret the change as a JSON encoded string",
    )
    type_group.add_argument(
        "-e",
        "--environment",
        dest="change_type",
        action="store_const",
        const="environment",
        required=False,
        help="Tells the script to interpret the change as an environment variable storing the JSON "
        + "encoded change.",
    )

    # Update Mode
    mode_group = parser.add_mutually_exclusive_group()
    mode_group.add_argument(
        "-l",
        "--local",
        dest="run_local",
        action="store_true",
        required=False,
        help="Tells the script to run locally, local is the default mode.",
    )
    mode_group.add_argument(
        "-r",
        "--remote",
        dest="run_local",
        action="store_false",
        required=False,
        help="Tells the script to run remote using the remote component from the config.",
    )

    parser.set_defaults(change_type="file", run_local=True, func=run_command_main)


def run_command_main(args: Namespace) -> None:
    """The main method for the update command, handles the actual execution of an update.

    Args:
        args (Namespace): The arguments supplied to the update command, such as the change.
    """

    # pylint: disable=unspecified-encoding

    event_handler = EventHandler.get()
    if args.verbose:
        event_handler.set_logging_level(LoggingLevel.VERBOSE)
    if args.debug:
        event_handler.set_logging_level(LoggingLevel.DEBUG)
    change = args.change
    event_handler.handle(DebugEvent({"message": f"Change: ({args.change_type}) {args.change}"}))
    event_args = {"change": args.change, "change_type": args.change_type}
    if args.change_type == "file":
        with open(change, "r") as change_file:
            change = change_factory.get_instance(json.loads(change_file.read()))
    elif args.change_type == "environment":
        change = os.getenv(change)
        assert isinstance(change, str)
        change = change_factory.get_instance(json.loads(change))
    else:
        change = change_factory.get_instance(json.loads(change))

    if args.change_type != "string":
        event_handler.handle(VerboseEvent({"message": f"Change: {change!r}"}))

    if args.run_local:
        event_handler.handle(VerboseEvent({"message": "Running locally"}))
        event_args["remote"] = False
        config_runner = get_config().local_runner
        if config_runner is None:
            event_handler.handle(DebugEvent({"message": "No runner defined, using default"}))
            runner: Runner = LocalRunner()
        else:
            runner = config_runner
    else:
        event_handler.handle(VerboseEvent({"message": "Running remote"}))
        event_args["remote"] = True
        config_runner = get_config().remote_runner
        assert config_runner is not None
        runner = config_runner

    event_args["runner"] = json.dumps(runner.bundle())
    event_handler.handle(ScriptRunEvent({"script": "update", "args": event_args}))
    runner.update(change)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""All Actions associated with a Change's labels."""

from typing import ClassVar, List

from pydantic import validator

from autotransform.change.base import Change
from autotransform.step.action.base import Action, ActionName


class AddLabelsAction(Action):
    """Adds labels to an existing Change.

    Attributes:
        labels(List[str]): The list of labels to add.
        name (ClassVar[ActionName]): The name of the component.
    """

    labels: List[str]

    name: ClassVar[ActionName] = ActionName.ADD_LABELS

    # pylint: disable=invalid-name
    @validator("labels")
    @classmethod
    def labels_must_be_non_empty(cls, v: List[str]) -> List[str]:
        """Validates the labels are not empty.

        Args:
            v (List[str]): The labels to add to the Change.

        Raises:
            ValueError: Raises an error when the labels are empty.

        Returns:
            List[str]: The unmodified labels to add.
        """

        if not v:
            raise ValueError("At least 1 label must be provided")
        if any(label == "" for label in v):
            raise ValueError("Labels must be non-empty strings")
        return v

    def run(self, change: Change) -> bool:
        """Adds labels to the specified Change.

        Args:
            change (Change): The Change to add labels to.

        Returns:
            bool: Whether the labels were added successful.
        """

        return change.add_labels(self.labels)


class RemoveLabelAction(Action):
    """Removes a label from an existing Change.

    Attributes:
        label(str): The label to remove.
        name (ClassVar[ActionName]): The name of the component.
    """

    label: str

    name: ClassVar[ActionName] = ActionName.REMOVE_LABEL

    # pylint: disable=invalid-name
    @validator("label")
    @classmethod
    def label_must_be_non_empty(cls, v: str) -> str:
        """Validates the label is not empty.

        Args:
            v (str): The label to remove.

        Raises:
            ValueError: Raises an error when the label is empty.

        Returns:
            str: The unmodified label of the comment.
        """

        if v == "":
            raise ValueError("Label to remove must be non-empty")
        return v

    def run(self, change: Change) -> bool:
        """Removes labels from the specified Change.

        Args:
            change (Change): The Change to remove labels from.

        Returns:
            bool: Whether the labels were removed successful.
        """

        return change.remove_label(self.label)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the RegexTransformer."""

from __future__ import annotations

import re
from typing import ClassVar

from autotransform.item.base import Item
from autotransform.item.file import FileItem
from autotransform.transformer.base import TransformerName
from autotransform.transformer.single import SingleTransformer


class RegexTransformer(SingleTransformer):
    """A Transformer that makes regex based changes, replacing instances of the provided
    pattern with the provided replacement. Useful for simple transformations. Operates on
    FileItems.

    Attributes:
        pattern (str): The pattern to search for.
        replacement (str): The value to replace the pattern with.
        name (ClassVar[TransformerName]): The name of the component.
    """

    pattern: str
    replacement: str

    name: ClassVar[TransformerName] = TransformerName.REGEX

    def _transform_item(self, item: Item) -> None:
        """Replaces all instances of a pattern in the file with the replacement string.

        Args:
            item (Item): The file that will be transformed.
        """

        # pylint: disable=unspecified-encoding

        assert isinstance(item, FileItem)
        content = item.get_content()
        new_content = re.sub(self.pattern, self.replacement, content)
        item.write_content(new_content)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""A utility for performing REST API requests."""

import json
import re
from functools import cached_property
from typing import Any, Callable, Dict, Mapping

import requests
from autotransform.event.debug import DebugEvent
from autotransform.event.handler import EventHandler
from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module


class RequestHandler(BaseModel):
    """Performs REST API requests, replacing values with things like environment variables

    Attributes:
        url(str): The URL to send a request to.
        constant_ replacers (optional, Mapping[str, Callable[[str], str]]): The replacers
            to apply to all usages of this handler. Defaults to {}.
        data(optional, Mapping[str, Any]): Data to include in the request. Defaults to {}.
        headers(optional, Mapping[str, Any]): Headers to include in the request. Defaults to {}.
        log_response(optional, bool): Indicates whether to log the response using DebugEvent.
            Defaults to False.
        params(optional, Mapping[str, Any]): Params to include in the request. Defaults to {}.
        post(optional, bool): Whether to send the request as a POST. Defaults to True.
    """

    url: str

    constant_replacers: Mapping[str, Callable[[str], str]] = Field(default_factory=dict)
    data: Mapping[str, Any] = Field(default_factory=dict)
    headers: Mapping[str, Any] = Field(default_factory=dict)
    log_response: bool = False
    params: Mapping[str, Any] = Field(default_factory=dict)
    post: bool = True

    @cached_property
    def _headers(self) -> Dict[str, Any]:
        """Gets the headers with any constant replacers filled in.

        Returns:
            Dict[str, Any]: The headers with constant replacers filled in.
        """

        headers = dict(self.headers)
        for name, replacer in self.constant_replacers.items():
            headers = self.replace_values(headers, name, replacer)

        return headers

    @cached_property
    def _params(self) -> Dict[str, Any]:
        """Gets the params with any constant replacers filled in.

        Returns:
            Dict[str, Any]: The params with constant replacers filled in.
        """

        params = dict(self.params)
        for name, replacer in self.constant_replacers.items():
            params = self.replace_values(params, name, replacer)

        return params

    @cached_property
    def _data(self) -> Dict[str, Any]:
        """Gets the data with any constant replacers filled in.

        Returns:
            Dict[str, Any]: The data with constant replacers filled in.
        """

        data = dict(self.data)
        for name, replacer in self.constant_replacers.items():
            data = self.replace_values(data, name, replacer)

        return data

    @staticmethod
    def replace_values(
        data: Mapping[str, Any], identifier: str, replacer: Callable[[str], str]
    ) -> Dict[str, Any]:
        """Replaces values in a dictionary with values from a replacing function.

        Args:
            data (Mapping[str, Any]): The data to replace values for.
            identifier (str): The identifier of the type of value.
            replacer (Callable[[str], Any]): The replacing function.

        Returns:
            Dict[str, Any]: The replaced data.
        """

        replaced_data: Dict[str, Any] = {}
        for name, val in data.items():
            if isinstance(val, str):
                match = re.match(f"<{identifier}:([^>]+)>", val)
                if match is not None:
                    tmp_val = val
                    for group in match.groups():
                        tmp_val = re.sub(f"<{identifier}:{group}>", replacer(group), tmp_val)
                    replaced_data[name] = tmp_val
                    continue
            elif isinstance(val, Mapping):
                replaced_data[name] = RequestHandler.replace_values(val, identifier, replacer)
                continue
            replaced_data[name] = val

        return replaced_data

    def get_response(self, replacers: Dict[str, Callable[[str], str]]) -> requests.Response:
        """Gets the value from a REST API request.

        Args:
            replacers (Dict[str, Callable[[str], str]]): The replacers to use for the request.

        Returns:
            requests.Response: The response to the request.
        """

        event_handler = EventHandler.get()

        headers = self._headers
        params = self._params
        data = self._data
        for name, replacer in replacers.items():
            headers = self.replace_values(headers, name, replacer)
            params = self.replace_values(params, name, replacer)
            data = self.replace_values(data, name, replacer)

        if self.post:
            response = requests.post(
                self.url, headers=headers, params=params, data=json.dumps(data), timeout=120
            )
        else:
            response = requests.get(
                self.url, headers=headers, params=params, data=json.dumps(data), timeout=120
            )

        if self.log_response:
            try:
                message = json.dumps(response.json(), indent=4)
            # pylint: disable=broad-except
            except Exception:
                message = str(response.content, encoding=response.apparent_encoding)

            event_handler.handle(DebugEvent({"message": f"Response:\n{message}"}))

        return response



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Filter components."""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import ClassVar, Optional, Sequence, Set

from autotransform.item.base import Item
from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent


class FilterName(str, Enum):
    """A simple enum for mapping."""

    AGGREGATE = "aggregate"
    CODEOWNERS = "codeowners"
    REGEX = "regex"
    REGEX_FILE_CONTENT = "regex_file_content"
    SCRIPT = "script"

    # Shard Filters
    KEY_HASH_SHARD = "key_hash_shard"


class Filter(NamedComponent):
    """The base for Filter components. Used by AutoTransform to determine if an Item from an Input
    is eligible for transformation.

    Attributes:
        inverted (bool, optional): Whether to invert the results of the filter. Defaults to False.
        name (ClassVar[FilterName]): The name of the component.
    """

    inverted: bool = False

    name: ClassVar[FilterName]

    def is_valid(self, item: Item) -> bool:
        """Check whether an Item is valid based on the Filter and handle inversion.

        Args:
            item (Item): The Item to check.

        Returns:
            bool: Returns True if the Item is eligible for transformation.
        """

        return self.inverted != self._is_valid(item)

    @abstractmethod
    def _is_valid(self, item: Item) -> bool:
        """Check whether an Item is valid based on the Filter. Does not handle inversion.

        Args:
            item (Item): The Item to check.

        Returns:
            bool: Returns True if the Item is eligible for transformation.
        """


class BulkFilter(Filter):
    """The base for BulkFilter components. Handles validation in bulk to determine valid Items.

    Attributes:
        _valid_keys (Optional[List[str]], optional): Whether to invert the results of the filter.
            Defaults to an None.
        name (ClassVar[FilterName]): The name of the component.
    """

    _valid_keys: Optional[Set[str]] = None

    @abstractmethod
    def _get_valid_keys(self, items: Sequence[Item]) -> Set[str]:
        """Gets the valid keys from the Items.

        Args:
            items (Sequence[Item]): The Items to check for valid items.

        Returns:
            Set[str]: The keys of the valid Items.
        """

    def pre_process(self, items: Sequence[Item]) -> None:
        """Sets up the _valid_keys set for the Filter.

        Args:
            items (Sequence[Item]): The Items to validate.
        """

        if self._valid_keys is None:
            self._valid_keys = self._get_valid_keys(items)


    def _is_valid(self, item: Item) -> bool:
        """Check whether an Item is valid based on the Filter. Does not handle inversion.

        Args:
            item (Item): The Item to check.

        Returns:
            bool: Returns True if the Item is eligible for transformation.
        """

        return self._valid_keys is not None and item.key in self._valid_keys


FACTORY = ComponentFactory(
    {
        FilterName.AGGREGATE: ComponentImport(
            class_name="AggregateFilter", module="autotransform.filter.aggregate"
        ),
        FilterName.CODEOWNERS: ComponentImport(
            class_name="CodeownersFilter", module="autotransform.filter.codeowners"
        ),
        FilterName.REGEX: ComponentImport(
            class_name="RegexFilter", module="autotransform.filter.regex"
        ),
        FilterName.REGEX_FILE_CONTENT: ComponentImport(
            class_name="RegexFileContentFilter", module="autotransform.filter.regex"
        ),
        FilterName.SCRIPT: ComponentImport(
            class_name="ScriptFilter", module="autotransform.filter.script"
        ),
        FilterName.KEY_HASH_SHARD: ComponentImport(
            class_name="KeyHashShardFilter", module="autotransform.filter.key_hash_shard"
        ),
    },
    Filter,  # type: ignore [type-abstract]
    "filter.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Provides utility methods for interacting with the scheduling of AutoTransform."""

from __future__ import annotations

import json
import os
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Type

from autotransform.event.debug import DebugEvent
from autotransform.event.handler import EventHandler
from autotransform.event.schedulerun import ScheduleRunEvent
from autotransform.filter.base import FACTORY as filter_factory
from autotransform.filter.key_hash_shard import KeyHashShardFilter
from autotransform.filter.shard import ShardFilter
from autotransform.runner.base import Runner
from autotransform.util.component import ComponentModel
from autotransform.util.console import (
    choose_option,
    choose_options_from_list,
    choose_yes_or_no,
    error,
    get_str,
    info,
    input_int,
    input_ints,
)
from autotransform.util.schema_map import SchemaMap
from pydantic import validator


class RepeatSetting(str, Enum):
    """Possible repitition schedules."""

    DAILY = "daily"
    WEEKLY = "weekly"


class SchemaScheduleSettings(ComponentModel):
    """A schedule for a Schema.

    Attributes:
        repeats (RepeatSetting): How often the Schema will be run, daily or weekly.
        hour_of_day (int, optional): The hour of the day to run the Schema, 0-23. Defaults to 0.
        day_of_week (Optional[int], optional): The day of the week to run a weekly Schema.
            Defaults to 0.
        shard_filter (Optional[ShardingSettings], optional): A shard filter to use for sharding
            the inputs across multiple runs. Defaults to None.
    """

    repeats: RepeatSetting
    hour_of_day: int = 0
    day_of_week: Optional[int] = 0
    shard_filter: Optional[ShardFilter] = None

    def should_run(self, hour_of_day: int, day_of_week: int) -> bool:
        """Whether the schema should be run.

        Args:
            hour_of_day (int): The hour of day to check.
            day_of_week (int): The day of week to check.

        Returns:
            bool: Whether the schema should run.
        """

        return self.hour_of_day == hour_of_day and (
            self.repeats == RepeatSetting.DAILY or self.day_of_week == day_of_week
        )

    @staticmethod
    def from_data(data: Dict[str, Any]) -> SchemaScheduleSettings:
        """Produces an instance of the SchemaScheduleSettings from decoded data.

        Args:
            data (Mapping[str, Any]): The JSON decoded data.

        Returns:
            SchemaScheduleSettings: An instance of the SchemaScheduleSettings.
        """

        repeats = RepeatSetting(data["repeats"])
        hour_of_day = data.get("hour_of_day", 0)
        assert isinstance(hour_of_day, int)
        assert hour_of_day in range(24)
        day_of_week = data.get("day_of_week", 0)
        if day_of_week is not None:
            assert isinstance(day_of_week, int)
            assert day_of_week in range(7)

        shard_filter = data.get("shard_filter", None)
        if shard_filter is not None:
            shard_filter = filter_factory.get_instance(shard_filter)
        else:
            shard_filter = None

        return SchemaScheduleSettings(
            repeats=repeats,
            hour_of_day=hour_of_day,
            day_of_week=day_of_week,
            shard_filter=shard_filter,
        )

    @staticmethod
    def from_console(simple: bool = False) -> SchemaScheduleSettings:
        """Gets a SchemaScheduleSettings from console inputs.

        Attributes:
            simple (bool, optional): Whether to use simple settings. Defaults to False.

        Returns:
            SchemaScheduleSettings: The input SchemaScheduleSettings.
        """

        # Get repeats
        repeats = choose_option(
            "How often should the schema run?",
            [(RepeatSetting.DAILY, ["d"]), (RepeatSetting.WEEKLY, ["w"])],
        )

        # Get hour_of_day
        hour_of_day = input_int(
            "Enter the hour of the day to run the schema", min_val=0, max_val=23
        )

        # Get day_of_week
        if repeats == RepeatSetting.WEEKLY:
            day_of_week = input_int(
                "Enter the day of the week to run the schema", min_val=0, max_val=6
            )
        else:
            day_of_week = None

        # Get shard_filter
        if not simple and choose_yes_or_no("Do you want to use sharding for this schema?"):
            shard_filter = filter_factory.from_console(
                "shard filter",
                default_value=KeyHashShardFilter(num_shards=10),
                allow_none=False,
            )
            while not isinstance(shard_filter, ShardFilter):
                assert shard_filter is not None
                error(f"{shard_filter.name} is not a shard filter")
                shard_filter = filter_factory.from_console(
                    "shard filter",
                    allow_none=False,
                )
        else:
            shard_filter = None

        return SchemaScheduleSettings(
            repeats=repeats,
            hour_of_day=hour_of_day,
            day_of_week=day_of_week,
            shard_filter=shard_filter,
        )


class ScheduledSchema(ComponentModel):
    """A Schema that is scheduled for automatic runs.

    Attributes:
        schema_name (str): The Schema that is being scheduled.
        schedule (SchemaScheduleSettings): The settings used to determine when to run the Schema.
        max_submissions (Optional[int], optional): The max number of submissions from a given
            scheduled run. No limit if None. Defaults to None.
    """

    schema_name: str
    schedule: SchemaScheduleSettings

    max_submissions: Optional[int] = None

    # pylint: disable=invalid-name
    @validator("max_submissions")
    @classmethod
    def max_submissions_is_positive(cls: Type[ScheduledSchema], v: Optional[int]) -> Optional[int]:
        """Validates that max submissions is positive.

        Args:
            cls (Type[ScheduledSchema]): The ScheduledSchema class.
            v (int): The maximum number of submissions.

        Raises:
            ValueError: Raised if the maximum number of submissions is not positive.

        Returns:
            Optional[int]: The unmodified maximum number of submissions.
        """

        if v is not None and v < 1:
            raise ValueError(f"Maximum number of submissions must be positive, {v} provided")
        return v

    @staticmethod
    def from_data(data: Dict[str, Any]) -> ScheduledSchema:
        """Produces an instance of the ScheduledSchema from decoded data.

        Args:
            data (Mapping[str, Any]): The JSON decoded data.

        Returns:
            ScheduledSchema: An instance of the ScheduledSchema.
        """

        schema_name = data["schema_name"]
        assert isinstance(schema_name, str)
        schedule = SchemaScheduleSettings.from_data(data["schedule"])
        max_submissions = data.get("max_submissions", None)
        if max_submissions is not None:
            assert isinstance(max_submissions, int)
        return ScheduledSchema(
            schema_name=schema_name, schedule=schedule, max_submissions=max_submissions
        )

    @staticmethod
    def from_console() -> ScheduledSchema:
        """Gets a ScheduledSchema from console inputs.

        Returns:
            ScheduledSchema: The input ScheduledSchema.
        """

        schema_name = get_str("Enter the schema to schedule: ")

        if choose_yes_or_no("Would you like to limit the maximum number of submissions?"):
            max_submissions = input_int("Enter the maximum number of submissions", min_val=1)
        else:
            max_submissions = None

        return ScheduledSchema(
            schema_name=schema_name,
            schedule=SchemaScheduleSettings.from_console(),
            max_submissions=max_submissions,
        )


class Scheduler(ComponentModel):
    """The information and functionality required to schedule Schemas.

    Attributes:
        base_time (int): The base time to use when determining hour_of_day, day_of_week,
            and valid shards. Considered day 0, hour 0.
        excluded_days (List[int]): A list of days of the week to skip running Schemas.
        schemas (List[ScheduledSchema]): A list of Schemas to schedule.
    """

    base_time: int
    excluded_days: List[int]
    schemas: List[ScheduledSchema]

    def run(self, start_time: int, runner: Runner) -> None:
        """Runs the schedule.

        Args:
            start_time (int): The time that the schedule is run on.
        """

        elapsed_time = start_time - self.base_time

        elapsed_hours = int(elapsed_time / 60 / 60)
        elapsed_days, hour_of_day = divmod(elapsed_hours, 24)
        elapsed_weeks, day_of_week = divmod(elapsed_days, 7)

        EventHandler.get().handle(
            DebugEvent({"message": f"Running for hour {hour_of_day}, day {day_of_week}"})
        )
        EventHandler.get().handle(
            DebugEvent({"message": f"Elapsed days {elapsed_days}, weeks {elapsed_weeks}"})
        )

        if day_of_week in self.excluded_days:
            EventHandler.get().handle(
                DebugEvent({"message": f"Day {day_of_week} is excluded, skipping run"})
            )
            return

        for scheduled_schema in self.schemas:
            # Check if should run
            if not scheduled_schema.schedule.should_run(hour_of_day, day_of_week):
                EventHandler.get().handle(
                    DebugEvent(
                        {"message": f"Skipping run of schema: {scheduled_schema.schema_name}"}
                    )
                )
                continue
            schema = SchemaMap.get().get_schema(scheduled_schema.schema_name)
            shard_filter = scheduled_schema.schedule.shard_filter
            if shard_filter is not None:
                if scheduled_schema.schedule.repeats == RepeatSetting.DAILY:
                    shard_filter.valid_shard = elapsed_days % shard_filter.num_shards
                else:
                    shard_filter.valid_shard = (elapsed_days // 7) % shard_filter.num_shards
                EventHandler.get().handle(DebugEvent({"message": f"Sharding: {shard_filter!r}"}))
                schema.filters.append(shard_filter)
            if scheduled_schema.max_submissions is not None:
                schema.config.max_submissions = scheduled_schema.max_submissions
            EventHandler.get().handle(ScheduleRunEvent({"schema_name": schema.config.schema_name}))
            runner.run(schema)

    def write(self, file_path: str) -> None:
        """Writes the Scheduler to a file as JSON.

        Args:
            file_path (str): The file to write to.
        """

        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w+", encoding="UTF-8") as scheduler_file:
            scheduler_file.write(json.dumps(self.bundle(), indent=4))
            scheduler_file.flush()

    @staticmethod
    def read(file_path: str) -> Scheduler:
        """Reads a Scheduler from a JSON encoded file.

        Args:
            file_path (str): The path where the JSON for the Schedule is located.

        Returns:
            Scheduler: The Scheduler from the file.
        """

        with open(file_path, "r", encoding="UTF-8") as scheduler_file:
            scheduler_json = scheduler_file.read()
        EventHandler.get().handle(
            DebugEvent({"message": f"Scheduler: ({file_path})\n{scheduler_json}"})
        )
        return Scheduler.from_json(scheduler_json)

    @staticmethod
    def from_json(scheduler_json: str) -> Scheduler:
        """Builds a Scheduler from JSON encoded values.

        Args:
            scheduler_json (str): The JSON encoded Scheduler.
            start_time (int): The start time to use for setting up sharding/running.

        Returns:
            Scheduler: The Scheduler from the JSON.
        """

        return Scheduler.from_data(json.loads(scheduler_json))

    @staticmethod
    def from_data(data: Dict[str, Any]) -> Scheduler:
        """Produces an instance of the Scheduler from decoded data.

        Args:
            data (Mapping[str, Any]): The JSON decoded data.

        Returns:
            Scheduler: An instance of the Scheduler.
        """

        base_time = data["base_time"]
        assert isinstance(base_time, int)
        excluded_days = data["excluded_days"]
        assert isinstance(excluded_days, List)
        for day in excluded_days:
            assert isinstance(day, int)
            assert day in range(7)
        schemas = [ScheduledSchema.from_data(schema) for schema in data.get("schemas", [])]
        return Scheduler(base_time=base_time, excluded_days=excluded_days, schemas=schemas)

    @staticmethod
    def init_from_console(use_sample_schema: bool = False, simple: bool = False) -> Scheduler:
        """Gets a Scheduler using console input.

        Args:
            use_sample_schema (bool, optional): Whether to include the sample Schema. Defaults
                to False.
            simple (bool, optional): Whether to use simple settings. Defaults to False.

        Returns:
            Scheduler: The input Scheduler.
        """

        # Gets base time
        info("Using local time to establish a base")
        info("Midnight Monday local is day_of_week 0, hour_of_day 0")
        if not simple and choose_yes_or_no("Apply a modifier to local time for scheduling?"):
            base_modifier = input_int("Enter the modifier in seconds")
        else:
            base_modifier = 0

        # Gets excluded days
        if not simple and choose_yes_or_no("Should AutoTransform schedule runs on weekends?"):
            excluded_days = []
        else:
            excluded_days = [5, 6]

        # Gets Schemas
        if use_sample_schema:
            schemas = [
                ScheduledSchema(
                    schema_name="Black Format",
                    schedule=SchemaScheduleSettings(repeats=RepeatSetting.DAILY, hour_of_day=7),
                )
            ]
        else:
            schemas = []
        if not simple:
            while choose_yes_or_no("Add a schema to the schedule?"):
                schemas.append(ScheduledSchema.from_console())

        return Scheduler(
            base_time=int(datetime.fromisoformat("2022-05-23T00:00:00").timestamp())
            + base_modifier,
            excluded_days=excluded_days,
            schemas=schemas,
        )

    @staticmethod
    def from_console(prev_scheduler: Optional[Scheduler] = None) -> Scheduler:
        """Gets a Scheduler using console inputs.

        Args:
            prev_scheduler (Optional[Scheduler], optional): A previously input Scheduler.
                Defaults to None.

        Returns:
            Scheduler: The input Scheduler.
        """

        # Get Base Time
        if prev_scheduler and choose_yes_or_no(
            f"Use previous base time ({prev_scheduler.base_time})?"
        ):
            base_time = prev_scheduler.base_time
        else:
            base_time = int(datetime.fromisoformat("2022-05-23T00:00:00").timestamp())
            info("Using local time to establish a base")
            info("Midnight Monday local is day_of_week 0, hour_of_day 0")
            info(f"Associated timestamp is {base_time}")
            if choose_yes_or_no("Apply a modifier to local time for scheduling?"):
                base_time += input_int("Enter the modifier in seconds")

        # Get Excluded Days
        if prev_scheduler is not None and choose_yes_or_no(
            f"Use previous excluded days: {prev_scheduler.excluded_days!r}"
        ):
            excluded_days = prev_scheduler.excluded_days
        else:
            excluded_days = input_ints(
                "Enter excluded days", min_val=0, max_val=6, min_choices=0, max_choices=7
            )

        # Get Schemas
        if prev_scheduler is not None and bool(prev_scheduler.schemas):
            schemas = choose_options_from_list(
                "Choose schemas to keep",
                [(schema, f"{schema!r}") for schema in prev_scheduler.schemas],
                min_choices=0,
                max_choices=len(prev_scheduler.schemas),
            )
        else:
            schemas = []
        while choose_yes_or_no("Add a schema to the schedule?"):
            schemas.append(ScheduledSchema.from_console())
        return Scheduler(base_time=base_time, excluded_days=excluded_days, schemas=schemas)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The RemoteRunEvent is triggered whenever an AutoTransformSchema is triggered for a run
on remote infrastructure.
"""

from typing import TypedDict

from autotransform.event.base import Event
from autotransform.event.logginglevel import LoggingLevel
from autotransform.event.type import EventType


class RemoteRunEventData(TypedDict):
    """The data for a RemoteRunEvent. Contains the information that will be
    logged when the event is triggered."""

    schema_name: str
    ref: str


class RemoteRunEvent(Event[RemoteRunEventData]):
    """A RemoteRunEvent is triggered whenever an AutoTransformSchema is triggered for a run
    on remote infrastructure.
    """

    @staticmethod
    def get_type() -> EventType:
        """Used to represent the type of Event, output to logs.

        Returns:
            EventType: The unique type associated with this Event.
        """

        return EventType.REMOTE_RUN

    @staticmethod
    def get_logging_level() -> LoggingLevel:
        """The logging level for events of this type.

        Returns:
            LoggingLevel: The logging detail required to log this event.
        """

        return LoggingLevel.INFO

    def _get_message(self) -> str:
        """Gets a message representing the details of the event.

        Returns:
            str: The message for the event.
        """

        return f"Remote run of {self.data['schema_name']}: {self.data['ref']}"



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Batchers take filtered Items and separate them in to logical groupings
with associated metadata that can be acted on independently. Each Batch
represents a single Change (i.e. a Pull Request)."""



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""All Actions associated with a Change's reviewers."""

from typing import Any, ClassVar, Dict, List

from pydantic import Field, root_validator, validator

from autotransform.change.base import Change
from autotransform.step.action.base import Action, ActionName


class AddReviewersAction(Action):
    """Adds reviewers to an existing Change.

    Attributes:
        reviewers(optional, List[str]): The list of reviewers to add. Defaults to [].
        team_reviewers(optional, List[str]): The list of team reviewers to add. Defaults to [].
        name (ClassVar[ActionName]): The name of the component.
    """

    reviewers: List[str] = Field(default_factory=list)
    team_reviewers: List[str] = Field(default_factory=list)

    name: ClassVar[ActionName] = ActionName.ADD_REVIEWERS

    # pylint: disable=invalid-name
    @validator("reviewers")
    @classmethod
    def labels_must_be_non_empty(cls, v: List[str]) -> List[str]:
        """Validates the reviewers are not empty strings.

        Args:
            v (List[str]): The reviewers to add to the Change.

        Raises:
            ValueError: Raises an error if a reviewer is an empty string.

        Returns:
            List[str]: The unmodified reviewers to add.
        """

        if any(reviewer == "" for reviewer in v):
            raise ValueError("Reviewers must be non-empty strings")
        return v

    # pylint: disable=invalid-name
    @validator("team_reviewers")
    @classmethod
    def team_reviewers_must_be_non_empty(cls, v: List[str]) -> List[str]:
        """Validates the team reviewers are not empty strings.

        Args:
            v (List[str]): The team reviewers to add to the Change.

        Raises:
            ValueError: Raises an error if a team reviewer is an empty string.

        Returns:
            List[str]: The unmodified team reviewers to add.
        """

        if any(team_reviewer == "" for team_reviewer in v):
            raise ValueError("Team reviewers must be non-empty strings")
        return v

    @root_validator
    @classmethod
    def some_reviewers_must_be_provided(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Checks that either reviewers or team reviewers are supplied.

        Args:
            values (Dict[str, Any]): The values for the action.

        Raises:
            ValueError: Raises an error if neither reviewers nor team reviewers are supplied.

        Returns:
            Dict[str, Any]: The unmodified values.
        """

        reviewers, team_reviewers = values.get("reviewers"), values.get("team_reviewers")
        if not reviewers and not team_reviewers:
            raise ValueError("Either reviewers or team reviewers must be supplied")
        return values

    def run(self, change: Change) -> bool:
        """Adds reviewers to the specified Change.

        Args:
            change (Change): The Change to add reviewers to.

        Returns:
            bool: Whether the reviewers were added successful.
        """

        return change.add_reviewers(self.reviewers, self.team_reviewers)


class AddOwnersAsReviewersAction(Action):
    """Adds the owners of a Schema as reviewers for the Change.

    Attributes:
        name (ClassVar[ActionName]): The name of the component.
    """

    name: ClassVar[ActionName] = ActionName.ADD_OWNERS_AS_REVIEWERS

    def run(self, change: Change) -> bool:
        """Adds owners as reviewers to the specified Change.

        Args:
            change (Change): The Change to add reviewers to.

        Returns:
            bool: Whether the reviewers were added successful.
        """

        return change.add_reviewers(change.get_schema().config.owners, [])


class AddOwnersAsTeamReviewersAction(Action):
    """Adds the owners of a Schema as team reviewers for the Change.

    Attributes:
        name (ClassVar[ActionName]): The name of the component.
    """

    name: ClassVar[ActionName] = ActionName.ADD_OWNERS_AS_TEAM_REVIEWERS

    def run(self, change: Change) -> bool:
        """Adds owners as team reviewers to the specified Change.

        Args:
            change (Change): The Change to add team reviewers to.

        Returns:
            bool: Whether the team reviewers were added successful.
        """

        return change.add_reviewers([], change.get_schema().config.owners)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The ScheduleRunEvent is triggered whenever an AutoTransformSchema is scheduled to be run
by the schedule command of the AutoTransform script.
"""

from typing import TypedDict

from autotransform.event.base import Event
from autotransform.event.logginglevel import LoggingLevel
from autotransform.event.type import EventType


class ScheduleRunEventData(TypedDict):
    """The data for a ScheduleRunEvent. Contains the information that will be
    logged when the event is triggered."""

    schema_name: str


class ScheduleRunEvent(Event[ScheduleRunEventData]):
    """A ScheduleRunEvent is triggered whenever an AutoTransformSchema is scheduled to be
    run by the schedule command and logs the schedule.
    """

    @staticmethod
    def get_type() -> EventType:
        """Used to represent the type of Event, output to logs.

        Returns:
            EventType: The unique type associated with this Event.
        """

        return EventType.SCHEDULE_RUN

    @staticmethod
    def get_logging_level() -> LoggingLevel:
        """The logging level for events of this type.

        Returns:
            LoggingLevel: The logging detail required to log this event.
        """

        return LoggingLevel.INFO

    def _get_message(self) -> str:
        """Gets a message representing the details of the event.

        Returns:
            str: The message for the event.
        """

        return f"Scheduling run of {self.data['schema_name']}"



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Steps are used by the Manager to determine how to handle outstanding Changes
from AutoTransform. Merging accepted changes, abandoning rejected ones, updating
stale one, are all examples of potential steps."""



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the DirectoryInput."""

from __future__ import annotations

import os
from functools import cached_property
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Sequence

from autotransform.input.base import Input, InputName
from autotransform.item.file import FileItem
from pydantic import root_validator


class DirectoryInput(Input):
    """An Input that lists all files recursively within a provided directory.

    Attributes:
        paths (List[str]): The paths of the directories to fetch all files within.
        name (ClassVar[InputName]): The name of the component.
    """

    paths: List[str]

    name: ClassVar[InputName] = InputName.DIRECTORY

    @root_validator(pre=True)
    @classmethod
    def path_legacy_setting_validator(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validates paths using legacy path setting.

        Args:
            values (Dict[str, Any]): The values used to configure the DirectoryInput.

        Raises:
            ValueError: Raised if both path and paths are supplied.

        Returns:
            Mapping[str, Any]: The fixed values.
        """

        if "path" in values:
            if "paths" in values:
                raise ValueError("Can not supply both path and paths for DirectoryInput")
            values["paths"] = set([str(values["path"])])
        values["paths"] = list(set(values["paths"]))
        return values

    @cached_property
    def _files(self) -> List[str]:
        """A cached list of files within the directory."""

        files = []
        for directory in self.paths:
            for path, _, dir_files in os.walk(Path(directory)):
                files.extend([f"{path}/{file}" for file in dir_files])
            # Clean up file paths for consistency of use
            files = [file.replace("\\", "/").removeprefix("./") for file in files]
        return files

    def get_items(self) -> Sequence[FileItem]:
        """Gets a list of files recursively contained within the path.

        Returns:
            Sequence[FileItem]: The eligible files for transformation.
        """

        return [FileItem(key=file) for file in self._files]



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Github related utilities."""

from __future__ import annotations

import time
from datetime import datetime, timedelta
from functools import cached_property
from typing import Any, Dict, List, Optional, Set, Tuple
from urllib.error import HTTPError

import pytz
from autotransform.config import get_config
from autotransform.event.debug import DebugEvent
from autotransform.event.handler import EventHandler
from autotransform.event.warning import WarningEvent
from autotransform.repo.git import GitRepo
from fastcore.basics import AttrDict  # type: ignore
from ghapi.all import GhApi, gh2date  # type: ignore


class GithubUtils:
    """A class for utilities used to interact with Github. Stores instances of objects to prevent
    excessive API usage.

    Attributes:
        __instances (Dict[str, GithubUtils]): A mapping of repo names to Util objects.
        _api (GhApi): The API object used to handle requests to Github.
        _fully_qualified_repo (str): The fully qualified name of the Github repo.
        _gists (Dict[str, Gist]): A mapping from id to Gist.
        _pulls (Dict[int, PullRequest]): A mapping from pull number to PullRequest.
    """

    __instances: Dict[str, GithubUtils] = {}

    _api: GhApi
    _fully_qualified_repo: str
    _gists: Dict[str, Gist]
    _pulls: Dict[int, PullRequest]

    BEGIN_SCHEMA: str = "<<<<BEGIN SCHEMA>>>>"
    END_SCHEMA: str = "<<<<END SCHEMA>>>>"

    BEGIN_BATCH: str = "<<<<BEGIN BATCH>>>>"
    END_BATCH: str = "<<<<END BATCH>>>>"

    def __init__(self, fully_qualified_repo: str):
        """A simple constructor.

        Args:
            fully_qualified_repo (str): The fully qualified name of the repo, uses the
                format {owner}/{repo}.
        """

        assert fully_qualified_repo not in GithubUtils.__instances
        token = get_config().github_token
        url = get_config().github_base_url
        repo_parts = fully_qualified_repo.split("/")
        self._api = GhApi(token=token, gh_host=url, owner=repo_parts[0], repo=repo_parts[1])
        self._fully_qualified_repo = fully_qualified_repo
        self._gists: Dict[str, Gist] = {}
        self._pulls: Dict[int, PullRequest] = {}

    @staticmethod
    def get(fully_qualified_repo: str) -> GithubUtils:
        """Gets an instance of the GithubUtils for a specific repo which is stored in a cache.

        Args:
            fully_qualified_repo (str): The fully qualified name of the repo, uses the
                format {owner}/{repo}.

        Returns:
            GithubUtils: A util class with a cached API object for requests.
        """
        if fully_qualified_repo not in GithubUtils.__instances:
            util = GithubUtils(fully_qualified_repo)
            GithubUtils.__instances[fully_qualified_repo] = util
        return GithubUtils.__instances[fully_qualified_repo]

    def get_user_id(self) -> int:
        """Gets the user ID of the authenticated user.

        Returns:
            int: The user ID of the authenticated user.
        """

        return self._api.users.get_authenticated().id

    def create_pull_request(self, title: str, body: str, base: str, head: str) -> PullRequest:
        """Create a pull request with the given information.

        Args:
            title (str): The title of the pull request.
            body (str): The body of the pull request.
            base (str): The base branch the pull request is against.
            head (str): The head to apply to the base branch.

        Returns:
            PullRequest: The created pull request.
        """

        pull = self._api.pulls.create(title=title, head=head, base=base, body=body)
        return PullRequest(self._api, pull.number)

    def get_pull_request(self, pull_number: int) -> PullRequest:
        """Gets a pull request.

        Args:
            pull_number (int): The number of the pull request.

        Returns:
            PullRequest: The pull request.
        """

        if pull_number not in self._pulls:
            self._pulls[pull_number] = PullRequest(self._api, pull_number)
        return self._pulls[pull_number]

    def create_gist(
        self, files: Dict[str, Dict[str, str]], description: str, public: bool = True
    ) -> Gist:
        """Creates a Gist containing the supplied information

        Args:
            files (Dict[str, Dict[str, str]]): The files to include in the gist.
            description (str): A simple description of the gist.
            public (bool, optional): Whether the gist should be public. Defaults to True.

        Returns:
            Gist: The created gist.
        """

        res = self._api.gists.create(description=description, files=files, public=public)
        return Gist(self._api, res.id)

    def get_gist(self, gist_id: str) -> Gist:
        """Gets a wrapper around the requested gist.

        Args:
            gist_id (str): The id of the gist.

        Returns:
            Gist: The requested gist.
        """

        if gist_id not in self._gists:
            self._gists[gist_id] = Gist(self._api, gist_id)
        return self._gists[gist_id]

    def get_open_pull_requests(self, base: Optional[str] = None) -> List[PullRequest]:
        """Gets all outstanding pull requests from the repo.

        Args:
            base (Optional[str], optional): The base branch to fetch pull requests against.
                Defaults to None.

        Returns:
            List[PullRequest]: The list of all requests outstanding for the repo.
        """

        username = self._api.users.get_authenticated().login
        query = f"type:pr state:open author:{username} repo:{self._fully_qualified_repo}"
        if base is not None:
            query = f"{query} base:{base}"
        page = 1
        num_per_page = 100
        fetch_more = True
        all_pulls: Set[int] = set()
        while fetch_more:
            prs = self._api.search.issues_and_pull_requests(
                q=query, sort="created", order="desc", per_page=num_per_page, page=page
            )
            fetch_more = len(prs["items"]) == num_per_page
            page = page + 1
            all_pulls = all_pulls.union([pr.number for pr in prs["items"]])

        pulls = [self.get_pull_request(pull_number) for pull_number in all_pulls]

        return [pull for pull in pulls if pull.branch.startswith(GitRepo.BRANCH_NAME_PREFIX)]

    def create_workflow_dispatch(
        self, workflow: str | int, ref: str, inputs: Dict[str, Any]
    ) -> Optional[str]:
        """Creates a workflow dispatch event and runs it.

        Args:
            workflow (str | int): The ID or filename of the workflow to run.
            ref (str): The ref to run the workflow on.
            inputs (Dict[str, Any]): Any inputs the workflow needs.

        Returns:
            Optional[str]: The best guess URL of the workflow run. None if failed.
        """

        try:
            current_time = datetime.now().replace(microsecond=0)
            check_time = current_time - timedelta(days=1)
            self._api.actions.create_workflow_dispatch(workflow_id=workflow, ref=ref, inputs=inputs)
            # We wait a bit to make sure Github's API is updated before printing a best guess of the
            # Workflow run's URL
            time.sleep(2)
            EventHandler.get().handle(DebugEvent({"message": "Checking for workflow run URL"}))
            res = self._api.actions.list_workflow_runs(
                workflow_id=workflow,
                actor=self._api.users.get_authenticated().login,
                branch=ref,
                event="workflow_dispatch",
                created=f">={check_time.isoformat()}",
            )
            if not res.workflow_runs:
                return ""
            return res.workflow_runs[0].html_url
        except HTTPError as err:
            EventHandler.get().handle(
                WarningEvent({"message": f"Failed dispatch workflow {workflow}: {err}"})
            )
            return None


# pylint: disable=too-many-public-methods
class PullRequest:
    """A wrapper around GhApi pull request response for simplified access to pull request info.

    Attributes:
        number (int): The number of the pull request.
        _api (GhApi): The API object used to access Github's API.
    """

    number: int

    _api: GhApi

    def __init__(self, api: GhApi, pull_number: int):
        self.number = pull_number
        self._api = api

    @property
    def body(self) -> str:
        """Gets the body of the pull request.

        Returns:
            str: The body of the pull request.
        """

        return self._detailed_info.body

    @property
    def branch(self) -> str:
        """Gets the name of the head branch for the pull request.

        Returns:
            str: The name of the head branch.
        """

        return self._detailed_info.head.ref

    @property
    def merged(self) -> bool:
        """Whether the pull request has been merged.

        Returns:
            bool: Whether the pull request was merged.
        """

        return self._detailed_info.merged

    @property
    def owner_id(self) -> int:
        """The ID of the user that created the pull request.

        Returns:
            int: The ID of the owner of the pull request.
        """

        return self._detailed_info.user.id

    def is_open(self) -> bool:
        """A simple check if the pull request is open.

        Returns:
            bool: Whether the pull request is open.
        """

        return self._detailed_info.state == "open"

    def is_closed(self) -> bool:
        """A simple check if the pull request is closed.

        Returns:
            bool: Whether the pull request is closed.
        """

        return not self.is_open()

    def get_review_state(self) -> Optional[str]:
        """The state of the most recent review, only checking reviews that are
        "APPROVED" or "CHANGES_REQUESTED". If there are no approving or change_requested
        reviews, returns None.

        Returns:
            Optional[str]: The state of the review.
        """

        review_states = ["APPROVED", "CHANGES_REQUESTED"]
        reviews = self._api.pulls.list_reviews(pull_number=self.number)
        reviews.reverse()
        for review in reviews:
            if review.state in review_states:
                return review.state
        return None

    def get_test_state(self) -> Optional[str]:
        """Gets the state of tests on the PR using the checks API.

        Returns:
            Optional[str]: The combined state of test runs.
        """

        possible_conclusions = [
            "action_required",
            "canceled",
            "timed_out",
            "failure",
            "neutral",
            "success",
        ]

        checks = self._api.checks.list_for_ref(ref=self.branch)
        for check in checks.check_runs:
            if check.status != "completed":
                return "pending"
        results = {run.conclusion for run in checks.check_runs}
        for conclusion in possible_conclusions:
            if conclusion in results:
                return conclusion
        return "success"

    def get_labels(self) -> List[str]:
        """Gets the labels for a Pull Request.

        Returns:
            List[str]: The labels for a Pull Request.
        """

        labels = self._api.issues.list_labels_on_issue(issue_number=self.number)
        return [label["name"] for label in labels]

    def get_reviewers(self) -> List[str]:
        """Gets the requested reviewers for a Pull Request.

        Returns:
            List[str]: The requested reviewers for a Pull Request.
        """

        return self._reviewers[0]

    def get_team_reviewers(self) -> List[str]:
        """Gets the requested team reviewers for a Pull Request.

        Returns:
            List[str]: The requested team reviewers for a Pull Request.
        """

        return self._reviewers[1]

    @cached_property
    def _detailed_info(self) -> AttrDict:
        """A cached value of the detailed info from the Github API.

        Returns:
            AttrDict: The detailed info of the PR.
        """

        return self._api.pulls.get(pull_number=self.number)

    @cached_property
    def _reviewers(self) -> Tuple[List[str], List[str]]:
        """A cached value of the reviewers of the change.

        Returns:
            Tuple[List[str], List[str]]: A tuple containing user reviewers and team reviewers.
        """

        requested_reviewers = self._api.pulls.list_requested_reviewers(pull_number=self.number)
        per_page = 50
        page = 1
        user_reviewers = [reviewer["login"] for reviewer in requested_reviewers["users"]]
        team_reviewers = [reviewer["name"] for reviewer in requested_reviewers["teams"]]
        more_reviewers = True
        while more_reviewers:
            reviews = self._api.pulls.list_reviews(
                pull_number=self.number, page=page, per_page=per_page
            )
            user_reviewers.extend(review["user"]["login"] for review in reviews)
            more_reviewers = len(reviews) == per_page
        return (list(set(user_reviewers)), list(set(team_reviewers)))

    def get_created_at(self) -> int:
        """Gets the created at timestamp.

        Returns:
            int: The timestamp for when the pull request was created.
        """
        return int(pytz.utc.localize(gh2date(self._detailed_info.created_at)).timestamp())

    def get_updated_at(self) -> int:
        """Gets the updated at timestamp.

        Returns:
            int: The timestamp for when the pull request was updated.
        """
        return int(pytz.utc.localize(gh2date(self._detailed_info.updated_at)).timestamp())

    def get_mergeable_state(self) -> str:
        """Gets the mergeable state of the PR.

        Returns:
            str: The mergeable state.
        """
        return (
            str(self._detailed_info.mergeable_state)
            if "mergeable_state" in self._detailed_info
            else "unknown"
        )

    def add_labels(self, labels: List[str]) -> None:
        """Adds a list of labels to the pull request.

        Args:
            labels (List[str]): The labels to add to the pull request.
        """

        self._api.issues.add_labels(issue_number=self.number, labels=labels)

    def add_reviewers(self, reviewers: List[str], team_reviewers: List[str]) -> None:
        """Adds a list of labels to the pull request.

        Args:
            reviewers (List[str]): The reviewers to request for the pull request.
            team_reviewers (List[str]): The team reviewers to request for the pull request.
        """

        self._api.pulls.request_reviewers(
            pull_number=self.number,
            reviewers=list(set(reviewers)),
            team_reviewers=list(set(team_reviewers)),
        )

    def create_comment(self, body: str) -> None:
        """Adds a comment to the pull request.

        Args:
            body (str): The body of the comment.
        """

        self._api.issues.create_comment(issue_number=self.number, body=body)

    def merge(self) -> bool:
        """Merges the pull request.

        Returns:
            bool: Whether the pull request was successfully merged.
        """

        try:
            res = self._api.pulls.merge(pull_number=self.number)
            return res.merged
        except HTTPError as err:
            EventHandler.get().handle(
                WarningEvent({"message": f"Failed merge pull request {self.number}: {err}"})
            )
            return False

    def remove_label(self, label: str) -> None:
        """Removes a label from the pull request.

        Args:
            label (str): The label to remove from the pull request.
        """

        self._api.issues.remove_label(issue_number=self.number, name=label)

    def close(self) -> bool:
        """Closes the pull request.

        Returns:
            bool: Whether the update happened successfully.
        """

        try:
            res = self._api.pulls.update(pull_number=self.number, state="closed")
            return res.state == "closed"
        except HTTPError as err:
            EventHandler.get().handle(
                WarningEvent({"message": f"Failed to close pull request {self.number}: {err}"})
            )
            return False

    def delete_branch(self) -> bool:
        """Deletes the branch associated with the pull request.

        Returns:
            bool: Whether the deletion was successful.
        """

        try:
            self._api.delete_branch(self.branch)
            return True
        except HTTPError as err:
            EventHandler.get().handle(
                WarningEvent({"message": f"Failed to delete branch {self.branch}: {err}"})
            )
            return False


class Gist:
    """A wrapper around GhApi gist response for simplified access to gist info.

    Attributes:
        id (str): The id of the gist.
        _api (GhApi): The API object used to access Github's API.
    """

    gist_id: str

    _api: GhApi

    def __init__(self, api: GhApi, gist_id: str):
        self.gist_id = gist_id
        self._api = api

    @cached_property
    def _detailed_info(self) -> AttrDict:
        """A cached value of the detailed info from the Github API.

        Returns:
            AttrDict: The detailed info of the gist.
        """

        return self._api.gists.get(gist_id=self.gist_id)

    def get_description(self) -> str:
        """Gets the description of the gist.

        Returns:
            str: The description of the gist.
        """

        return self._detailed_info.description

    def get_file_content(self, file_name: str) -> Optional[str]:
        """Gets the content of the requested file within the gist.

        Args:
            file_name (str): The name of the file to get the content for.

        Returns:
            Optional[str]: The contents of the file. None if the gist does not have a file with
                the supplied name.
        """

        if file_name not in self._detailed_info.files:
            return None
        return self._detailed_info.files[file_name].content



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Events are used by AutoTransform for the purposes of logging. The EventHandler can be used
as an entry point for hooks in to the Event framework for logging to an organization's specific
datastore. Custom hooks can be added through the environment variable AUTO_TRANSFORM_EVENT_HANDLER.
Set that environment variable to a JSON map of:
{"class_name": <Class that inherits from EventHandler>, "module": <Module the class is in>}
to insert your own hook in to AutoTransform event handling.
"""



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the LocalRunner."""

from __future__ import annotations

from typing import ClassVar

from autotransform.change.base import Change
from autotransform.runner.base import Runner, RunnerName
from autotransform.schema.schema import AutoTransformSchema


class LocalRunner(Runner):
    """A Runner component that runs a Schema locally on the machine.

    Attributes:
        name (ClassVar[RunnerName]): The name of the component.
    """

    name: ClassVar[RunnerName] = RunnerName.LOCAL

    def run(self, schema: AutoTransformSchema) -> None:
        """Triggers a full run of a Schema locally.

        Args:
            schema (AutoTransformSchema): The schema that will be run.
        """

        schema.run()

    def update(self, change: Change) -> None:
        """Triggers an update of the Change.

        Args:
            change (Change): The Change to update.
        """

        schema = change.get_schema()
        batch = change.get_batch()
        schema.execute_batch(batch, change)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The EventHandler receives event dispatches and triggers appropriate behavior for
the event, such as logging. Set AUTO_TRANSFORM_EVENT_HANDLER environment variable to
{
    class_name: <The name of a class extending EventHandler>,
    module: <The fully qualified module where the class is>
}
as JSON to override the default event handling.
"""

from __future__ import annotations

import importlib
import json
import os
from datetime import datetime
from typing import Dict, Optional

from autotransform.event.base import Event
from autotransform.event.logginglevel import LoggingLevel
from colorama import Fore


class EventHandler:
    """The handler that all Events are dispatched to that logs these events.

    Attributes:
        _logging_level (LoggingLevel): The level for which logs will be output to CLI.
        __instance (Optional[EventHandler]): The singleton instance of the EventHandler.
        __color_map (Dict[LoggingLevel, str]): A mapping from log level to ANSI color for CLI
            output.
    """

    _logging_level: LoggingLevel
    __instance: Optional[EventHandler] = None
    __color_map: Dict[LoggingLevel, str] = {
        LoggingLevel.ERROR: Fore.RED,
        LoggingLevel.WARNING: Fore.YELLOW,
        LoggingLevel.INFO: Fore.WHITE,
        LoggingLevel.VERBOSE: Fore.GREEN,
        LoggingLevel.DEBUG: Fore.CYAN,
    }

    def __init__(self):
        if EventHandler.__instance is not None:
            # pylint: disable=broad-exception-raised
            raise Exception("Trying to instantiate new EventHandler when one already present")
        self._logging_level = LoggingLevel.INFO

    @staticmethod
    def get() -> EventHandler:
        """Singleton method for getting the event handler.

        Returns:
            EventHandler: The singleton instance of the EventHandler.
        """

        if EventHandler.__instance is None:
            event_handler_to_use = os.getenv("AUTO_TRANSFORM_EVENT_HANDLER")
            if event_handler_to_use is not None:
                try:
                    event_handler_info = json.loads(event_handler_to_use)
                    module = importlib.import_module(event_handler_info["module"])
                    event_handler = getattr(module, event_handler_info["class_name"])()
                    assert isinstance(event_handler, EventHandler)
                    EventHandler.__instance = event_handler
                except Exception:  # pylint: disable=broad-except
                    EventHandler.__instance = EventHandler()
            else:
                EventHandler.__instance = EventHandler()
        return EventHandler.__instance

    def set_logging_level(self, logging_level: LoggingLevel) -> None:
        """Sets the level of logs to include in console outputs.

        Args:
            logging_level (LoggingLevel): The logging level to output to console.
        """

        self._logging_level = logging_level

    def handle(self, event: Event) -> None:
        """Handles the given Event, logging and executing any hooks needed.

        Args:
            event (Event): The Event that was triggered.
        """

        if self._logging_level.value >= event.get_logging_level().value:
            self.output_to_cli(event)

    @staticmethod
    def output_to_cli(event: Event) -> None:
        """Outputs the event to CLI with appropriate coloring.

        Args:
            event (Event): The event being logged.
        """
        color = event.get_color_override()
        if color is None:
            color = EventHandler.__color_map[event.get_logging_level()]
        creation_string = datetime.fromtimestamp(event.create_time).strftime("%Y-%m-%d %H:%M:%S")
        print(f"{color}[{creation_string}]{event.get_message()}{Fore.RESET}")



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the GitRepo."""

from __future__ import annotations

import re
import subprocess
from functools import cached_property
from typing import Any, ClassVar, List, Mapping, Optional, Sequence

from git import Head
from git import Repo as GitPython

import autotransform.schema
from autotransform.batcher.base import Batch
from autotransform.change.base import Change
from autotransform.repo.base import Repo, RepoName


class GitRepo(Repo):
    """A Repo that provides support for commiting changes to git.

    Attributes:
        base_branch (str): The name of the base branch for the repository.
        name (ClassVar[str]): The name of the component.
        BRANCH_NAME_PREFIX (ClassVar[str]): The prefix to apply to branches that are created.
        COMMIT_MESSAGE_PREFIX (ClassVar[str]): The prefix to apply to commits that are created.
    """

    base_branch: str

    name: ClassVar[RepoName] = RepoName.GIT

    BRANCH_NAME_PREFIX: ClassVar[str] = "AUTO_TRANSFORM"
    COMMIT_MESSAGE_PREFIX: ClassVar[str] = "[AutoTransform]"

    @staticmethod
    def get_changed_files_from_status(status: str) -> List[str]:
        """Parses the git status to get the changed files.

        Args:
            status (str): The result of git status -s --untracked-files

        Returns:
            List[str]: The changed files.
        """

        if status.strip() == "":
            return []
        return [
            re.sub(r"^(?:\?\?|M|A|D)", "", line.strip()).strip()
            for line in status.strip().split("\n")
        ]

    @staticmethod
    def get_branch_name(title: str) -> str:
        """Gets a unique name for a git branch using the title from the Batch.

        Args:
            title (str): The title of the change.

        Returns:
            str: The name of the branch for this change.
        """

        # Handle titles of the format "[1/2] foo" that can come from chunk batching
        fixed_title = re.sub(r"\[(\d+)/(\d+)\]", r"\1_\2", title)

        if autotransform.schema.current is not None:
            schema_name = f"{autotransform.schema.current.config.schema_name}/"
        else:
            schema_name = ""

        branch_name = f"{GitRepo.BRANCH_NAME_PREFIX}/{schema_name}{fixed_title}"
        # Replace bad characters
        substring_replacements = {
            "/": "__",
            ":": "-",
            "^": "",
            "~": "",
            "..": "",
            " ": "_",
            "?": "",
            "*": "",
        }
        for substr, repl in substring_replacements.items():
            branch_name = branch_name.replace(substr, repl)

        return branch_name

    @staticmethod
    def get_commit_message(title: str) -> str:
        """Gets a commit message for the change based on the Batch title.

        Args:
            title (str): The title of the change.

        Returns:
            str: The commit message for this change.
        """

        # Add a blank space before prefixes
        if not title.startswith("["):
            title = f" {title}"
        if autotransform.schema.current is not None:
            schema_name = f"[{autotransform.schema.current.config.schema_name}]"
        else:
            schema_name = ""
        return f"{GitRepo.COMMIT_MESSAGE_PREFIX}{schema_name}{title}"

    @cached_property
    def _local_repo(self) -> GitPython:
        """Returns a cached instance of the local repo

        Returns:
            GitPython: The local repository.
        """

        dir_cmd = ["git", "rev-parse", "--show-toplevel"]
        repo_dir = subprocess.check_output(dir_cmd, encoding="UTF-8").replace("\\", "/").strip()
        return GitPython(repo_dir)

    @cached_property
    def _base_branch(self) -> Head:
        """Returns a cached instance of the base branch.

        Returns:
            Head: The base branch.
        """

        for branch in self._local_repo.heads:
            if branch.name == self.base_branch:
                return branch
        raise ValueError("Invalid base branch name, branch not found.")

    def get_changed_files(self, _batch: Batch) -> List[str]:
        """Uses git status to get all changed files.

        Args:
            _batch (Batch): Unused Batch object used to match signature to base.

        Returns:
            List[str]: All changed files, including untracked files.
        """

        return GitRepo.get_changed_files_from_status(
            self._local_repo.git.status("-s", untracked_files=True),
        )

    def has_outstanding_change(self, batch: Batch) -> bool:
        """Checks the state of the repo to see whether an outstanding Change
        for the Batch exists.

        Args:
            batch (Batch): The Batch to check for.

        Returns:
            bool: Whether an outstanding Change exists.
        """

        return GitRepo.get_branch_name(batch["title"]) in [
            str(ref) for ref in self._local_repo.references
        ]

    def submit(
        self,
        batch: Batch,
        _transform_data: Optional[Mapping[str, Any]],
        change: Optional[Change] = None,
    ) -> None:
        """Stages all changes and commits them in a new branch.

        Args:
            batch (Batch): The Batch for which the changes were made.
            _transform_data (Optional[Mapping[str, Any]]): Data from the transformation. Unused.
            change (Optional[Change]): An associated change which should be updated.
                Defaults to None.
        """

        self.commit(batch["title"], change is not None)

    def commit(self, title: str, update: bool) -> None:
        """Creates a new branch for all changes, stages them, and commits them.

        Args:
            title (str): The title of the Batch being commited.
            update(bool): Whether to update an existing change.
        """

        if update:
            self._local_repo.git.checkout("-B", GitRepo.get_branch_name(title))
        else:
            self._local_repo.git.checkout("-b", GitRepo.get_branch_name(title))
        self._local_repo.git.add(all=True)
        self._local_repo.index.commit(GitRepo.get_commit_message(title))

    def clean(self, _batch: Batch) -> None:
        """Performs `git reset --hard` to remove any changes.

        Args:
            _batch (Batch): Unused Batch object used to match signature to base
        """

        self._local_repo.git.reset("--hard")

    def rewind(self, batch: Batch) -> None:
        """First eliminates any uncommitted changes using the clean function then checks out
        the initial active branch.

        Args:
            batch (Batch): The Batch for the submitted changes that is being rewound.
        """

        self.clean(batch)
        self._base_branch.checkout()

    def get_outstanding_changes(self) -> Sequence[Change]:
        """Gets all outstanding Changes for the Repo.

        Returns:
            Sequence[Change]: The outstanding Changes against the Repo.
        """

        return []



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The EventType enum."""

from enum import Enum


class EventType(str, Enum):
    """A simple enum for Event naming"""

    DEBUG = "debug"
    MANAGE_ACTION = "manage_action"
    REMOTE_RUN = "remote_run"
    REMOTE_UPDATE = "remote_update"
    SCHEDULE_RUN = "schedule_run"
    SCRIPT_RUN = "script_run"
    VERBOSE = "verbose"
    WARNING = "warning"



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Provides configuration to AutoTransform. By default this configuration is pulled from
config files. The configuration fetcher used can be overridden using the AUTO_TRANSFORM_CONFIG
environment variable. This can support different methods of configuration, such as environment
variables or a custom approach where the environment variable is set to
"{'class_name': <A ConfigFetcher Class>, 'module': <The module where the class is located>}".
"""

from __future__ import annotations

import importlib
import json
import os
import subprocess
from functools import cache
from typing import TYPE_CHECKING, Dict, Optional

if TYPE_CHECKING:
    from autotransform.config.config import Config

CONFIG_FILE_NAME = "config.json"


def get_repo_config_relative_path() -> str:
    """Gets the relative path used for the repo config directory.

    Returns:
        str: The relative path to the config.
    """

    return os.getenv("AUTO_TRANSFORM_REPO_CONFIG_PATH", "autotransform")


def get_repo_config_dir() -> Optional[str]:
    """Gets the directory where a repo-specific AutoTransform config file would be located. None
    if git fails. The environment variable AUTO_TRANSFORM_REPO_CONFIG_PATH can be used to set
    the relative path to the config.json file for repo configs.

    Returns:
        Optional[str]: The directory of the repo config file. None if git fails.
    """

    try:
        dir_cmd = ["git", "rev-parse", "--show-toplevel"]
        repo_dir = subprocess.check_output(dir_cmd, encoding="UTF-8").replace("\\", "/").strip()
        return f"{repo_dir}/{get_repo_config_relative_path()}"
    except Exception:  # pylint: disable=broad-except
        return None


def get_cwd_config_dir() -> str:
    """Gets the directory where an AutoTransform config file would be located relative to the
    current working directory. The environment variable AUTO_TRANSFORM_CWD_CONFIG_PATH can be
    used to set the relative path to the config.json file for CWD configs.

    Returns:
        str: The directory of the cwd config file.
    """

    relative_path = os.getenv("AUTO_TRANSFORM_CWD_CONFIG_PATH", "autotransform")
    cwd = os.getcwd().replace("\\", "/")
    return f"{cwd}/{relative_path}"


@cache
def get_config() -> Config:
    """Gets the Config for AutoTransform.

    Returns:
        Config: The cached Config.
    """

    # pylint: disable=import-outside-toplevel
    from autotransform.config.default import DefaultConfigFetcher
    from autotransform.config.environment import EnvironmentConfigFetcher
    from autotransform.config.fetcher import ConfigFetcher, ConfigFetcherName

    fetcher_to_use = os.getenv("AUTO_TRANSFORM_CONFIG")

    fetchers: Dict[str, ConfigFetcher] = {
        ConfigFetcherName.DEFAULT: DefaultConfigFetcher(),
        ConfigFetcherName.ENVIRONMENT: EnvironmentConfigFetcher(),
    }

    if fetcher_to_use is not None:
        if fetcher_to_use in fetchers:
            fetcher: ConfigFetcher = fetchers[fetcher_to_use]
        else:
            try:
                fetcher_info = json.loads(fetcher_to_use)
                module = importlib.import_module(fetcher_info["module"])
                fetcher = getattr(module, fetcher_info["class_name"]).from_data(
                    fetcher_info.get("data", {})
                )
                assert isinstance(fetcher, ConfigFetcher)
            except Exception:  # pylint: disable=broad-except
                fetcher = DefaultConfigFetcher()
    else:
        fetcher = DefaultConfigFetcher()

    return fetcher.get_config()
